// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: bi_tree_type.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "bi_tree_type.h"
#include "util/general/die.h"

#include <ostream>
#include <sstream>

using namespace std;

// parser for newick formatted trees (bifurcating only)
namespace NEWICK_PARSER {

  void
  parse_error(const char* s){
    static const std::string h("newick bifurcating tree parser: ");
    pass_away((h+s).c_str());
  }


  enum {
    START='(',
    STOP=')',
    JOIN=',',
    LABELD=':',
    END=';'
  };


  bool
  is_symbol(const char c){
    switch(c){
    case START:
    case STOP:
    case JOIN:
    case LABELD:
    case END:    return true;
    default:     return false;
    }
  }


  inline
  void
  next_token(const char*& s){
    while(isspace(*s)) ++s;
  }


  void
  parse_label(const char*& s,
              bi_tree_node* b){
    if(! b->label.empty()) parse_error("malformed tree string (2)");

    const char* start(s);
    for(;;++s){
      if( is_symbol(*s) || isspace(*s) || *s == '\0'){
        b->label=std::string(start,s-start);
        next_token(s);
        return;
      }
    }
  }


  void
  parse_length(const char*& s,
               bi_tree_node* b){
    char* s_end(0);
    /// \todo error checking float conversion here:
    b->length=strtod(s,&s_end);
    if(s==s_end) parse_error("can't parse branch length to floating point");
    b->is_length_set=true;
    s=s_end;
    next_token(s);
  }


  void
  parse_label_length(const char*& s,
                     bi_tree_node* b){
    parse_label(s,b);
    b->is_length_set=false;
    if(*s==LABELD){
      next_token(++s);
      parse_length(s,b);
    }
  }


  bi_tree_node*
  parse_subtree(const char*& s){

    bi_tree_node* n(new bi_tree_node());
    if(*s==START){
      next_token(++s);
      n->child1=parse_subtree(s);
      n->child1->parent=n;

      if(*s!=JOIN) parse_error("expected ',' in paren block");
      next_token(++s);
      n->child2=parse_subtree(s);
      n->child2->parent=n;

      if(*s!=STOP) parse_error("expected ')' to close paren block");
      next_token(++s);
    }
    parse_label_length(s,n);
    return n;
  }


  bi_tree_node*
  parse(const char* s){
    next_token(s);
    bi_tree_node* n(parse_subtree(s));
    if(*s=='\0') parse_error("unexpected termination (possibly no closing semicolon)");
    if(*s!=END) parse_error("malformed tree string");
    return n;
  }
}


// copys a bi_tree from the root down
//
bi_tree_node*
copy_node(const bi_tree_node* b,
          bi_tree_node* copy_parent){
  bi_tree_node* a=new bi_tree_node;
  a->label=b->label;
  a->length=b->length;
  a->is_length_set=b->is_length_set;
  a->node_index=b->node_index;
  a->leaf_index=b->leaf_index;
  if(copy_parent) a->parent=copy_parent;
  if(b->child1){a->child1=copy_node(b->child1,a);}
  if(b->child2){a->child2=copy_node(b->child2,a);}
  return a;
}


void
bi_tree_type::
read_tree(const char* s){
  // set up regular tree:
  clear();
  root = NEWICK_PARSER::parse(s);
}


std::ostream& operator<<(std::ostream& os,
                         const bi_tree_node* b){

  if(b==0) return os << "0";

  if(b->child1 || b->child2){
    os << "(";
    if(b->child1) os << b->child1;
    if(b->child1 && b->child2) os << ",";
    if(b->child2) os << b->child2;
    os << ")";
  }
  os << b->label;
  if(b->is_length_set) os << ":" << b->length;
  return os;
}


std::ostream& operator<<(std::ostream& os,
                         const bi_tree_type& t){
  return os << t.root << ";";
}


static
void
index_tree(bi_tree_node* b,
           std::vector<bi_tree_node*>& v){

  if(b->label==""){
    ostringstream oss;
    oss << "node" << v.size();
    b->label=oss.str();
  }

  v.push_back(b);

  if(b->child1) index_tree(b->child1,v);
  if(b->child2) index_tree(b->child2,v);
}


void
bi_tree_sml_type::
setup_tree_index(){

  tree.clear();
  tree_index_map.clear();
  leaf.clear();
  leaf_index_map.clear();

  // cleanup:
  if(root->is_length_set){
    warning("Input tree root branch length ignored");
    root->is_length_set=false;
  }

  // now reorder and index things for fast id lookup
  index_tree(root,tree);

  for(unsigned i(0);i<n_nodes();++i){
    tree_index_map[tree[i]->label]=i;
    tree[i]->node_index=i;
    if( tree[i]->child1==0 && tree[i]->child2==0 ){
      leaf_index_map[tree[i]->label]=leaf.size();
      tree[i]->leaf_index=leaf.size();
      leaf.push_back(tree[i]);
    }
  }
}



static
void
clear_lengths_node(bi_tree_node* b,
                   const bool is_warn){

  if(b->is_length_set) {
    if(is_warn){
      warning("Input tree branch lengths ignored");
    }
    b->is_length_set=false;
  }

  if(b->child1) clear_lengths_node(b->child1,is_warn);
  if(b->child2) clear_lengths_node(b->child2,is_warn);
}


void
bi_tree_sml_type::
clear_lengths(const bool is_warn){
  clear_lengths_node(root,is_warn);
}
