// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: cat_manager.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "cat_manager.h"
#include "util/general/die.h"
#include "util/general/io_util.h"
#include "util/general/metatags.h"

#include <iostream>
#include <sstream>
#include <string>


cat_manager::
cat_manager(const cat_manager_options& cmo)
  : _opt(cmo), cat_index(SITE_CAT_TYPE::SIZE) {

  using namespace SITE_CAT_TYPE;

  if(cmo.n_site_cats[ROOT] != 1) die("invalid root_cat state");

  unsigned n_cats(1);
  for(unsigned i(0);i<SIZE;++i) n_cats *= cmo.n_site_cats[i];

  for(unsigned i(0);i<SIZE;++i) cat_index[i].resize(n_cats);

  for(unsigned i(0);i<n_cats;++i){
    // default to fully-crossed cat combinations (except for ROOT):
    cat_index[MUT_RATE][i]=i%cmo.n_site_cats[MUT_RATE];
    const unsigned tmpu(i/cmo.n_site_cats[MUT_RATE]);

    cat_index[MUT_MODEL][i]=tmpu%cmo.n_site_cats[MUT_MODEL];
    const unsigned tmpu2(tmpu/cmo.n_site_cats[MUT_MODEL]);

    cat_index[SEL_STRENGTH][i]=tmpu2%cmo.n_site_cats[SEL_STRENGTH];
    cat_index[SEL_MATRIX][i]=tmpu2/cmo.n_site_cats[SEL_STRENGTH];
  }

  // determine root cats based on assigned data:
  const unsigned ac(get_assigned_cat_size());

  // stupid convention right now is that any assigned cats map 1:1 to
  // the full site cat set:
  //
  if(ac){
    assert(ac==n_cats); //stupid convention broken!!

    for(unsigned i(0);i<n_cats;++i){
      cat_index[ROOT][i]=i;
    }
  }
}



static const char* const section_id = "cat_manager";
#ifndef NEXT_MODELFILE_VERSION
const char* const sdata_label = "is_any_assigned_cats";
#endif
const char* const catmap_label ="data_catmap";
const char* const end_label = "END";


void
cat_manager::
store_state(std::ostream& os) const {

  {
#ifndef NEXT_MODELFILE_VERSION
    os << section_id << " " << sdata_label << " " << (get_assigned_cat_size()>0) << "\n";
#endif
    data_catmap_type::const_iterator i=data_catmap().begin(),i_end=data_catmap().end();
    for(;i!=i_end;++i){
      os << section_id << " " << catmap_label << " " << i->first << " " << i->second << "\n";
    }
  }

  const unsigned n_cats(get_cat_size());
  os << section_id << " n_site_cats " << n_cats << "\n";

  for(unsigned i(0);i<n_cats;++i){
    os << section_id << " site_cat_" << i << "_mr_mm_ss_sm_ro";
    for(unsigned j(0);j<SITE_CAT_TYPE::SIZE;++j){
      os << " " << cat_index[j][i];
    }
    os << "\n";
  }
  os << section_id << " " << end_label << "\n";
}


static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr)  NORETURN_TAG;

static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr){

  std::cerr << "ERROR:: -- Invalid model file format. --\n"
            << "ERROR:: expected_section: " << sid << "\n"
            << "ERROR:: section: " << tmpstr << "\n"
            << "\n";
  std::cerr << is.rdbuf();
  abort();
}



static
void
load_advance(std::istream& is,
             const char* const second = 0){

  std::string tmpstr;
  is >> tmpstr;
  if(tmpstr != section_id) load_die(is,section_id,tmpstr);
  is >> tmpstr;
  if(second && tmpstr != second) load_die(is,second,tmpstr);
}


cat_manager::
cat_manager(std::istream& is)
  : _opt(), cat_index(SITE_CAT_TYPE::SIZE) {

  unsigned n_cats;

  // clear out the end of a previous line:
  clear_whitespace_line(is);
  std::string buf;
  while(getline(is,buf)){
    std::istringstream sbuf(buf);

    std::string tmpstr;
    sbuf >> tmpstr;
    if(tmpstr != section_id){
      std::cerr << "ERROR:: -- Invalid model file format. --\n"
                << "ERROR:: expected_section: " << section_id << "\n"
                << "ERROR:: section: " << tmpstr << "\n"
                << "ERROR:: line buffer: " << buf << "\n";
      abort();
    }

    sbuf >> tmpstr;

    unsigned tmpu;
#ifndef NEXT_MODELFILE_VERSION
    if       (tmpstr == sdata_label){
      sbuf >> tmpu;
    } else
#endif
    if       (tmpstr == catmap_label){
      sbuf >> tmpstr >> tmpu;
      _opt.data_catmap[tmpstr] = tmpu;
    } else if(tmpstr == "n_site_cats"){
      sbuf >> n_cats;
      for(unsigned i(0);i<SITE_CAT_TYPE::SIZE;++i){
        cat_index[i].resize(n_cats);
      }
      break;
    } else {
      std::cerr << "ERROR:: invalid model file format. section: " << section_id << " subsection: " << tmpstr << "\n";
      abort();
    }
  }

  for(unsigned i(0);i<n_cats;++i){
    load_advance(is);
    for(unsigned j(0);j<SITE_CAT_TYPE::SIZE;++j){
      is >> cat_index[j][i];
    }
  }

  load_advance(is,end_label);
};
