// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: prob_gtor_obs_derived.h 741 2007-08-14 02:39:19Z ctsa $

/// \file
///
/// \brief 'parameterless' root distros derived from the observed
/// state dist
///


#ifndef __PROB_GTOR_OBS_DERIVED_H
#define __PROB_GTOR_OBS_DERIVED_H


#include "prob_gtor.h"
#ifdef BRANCH_SPECIFIC_OBS
#include "simple_util.h"
#endif


/// \brief average between the different org obs distros by tree
/// branch length
///
struct prob_gtor_obs_state_time_avg : prob_gtor<prob_t> {

  typedef prob_gtor_obs_state_time_avg self_t;
  typedef prob_gtor<prob_t> base_t;

  explicit
  prob_gtor_obs_state_time_avg(const root_gtor_sml_share& smlm,
                               const unsigned root_cat_no)
    : base_t(smlm.state_size(),0), _smlm(smlm), is_assigned_cat(false), assigned_cat_no(0) {

    // convention is all assigned cats are 1:1 mapped to root:
    if(_smlm.get_cat_manager().get_assigned_cat_size()){
      is_assigned_cat=true;
      assigned_cat_no=root_cat_no;
    }
  }

  prob_gtor_obs_state_time_avg(const self_t& s,
                               const root_gtor_sml_share& smlm)
    : base_t(s), _smlm(smlm), is_assigned_cat(s.is_assigned_cat), assigned_cat_no(0) {}

  virtual self_t* clone(const root_gtor_sml_share& smlm) const { return new self_t(*this,smlm); }

private:
  virtual void get_param_from_distro() const {}

  virtual void param_update(const bool is_secondary_update){
    if(is_secondary_update) _smlm.get_obs_state_time_avg_root(_distro,is_assigned_cat,assigned_cat_no);
  }

  const root_gtor_sml_share _smlm;
  bool is_assigned_cat;
  unsigned assigned_cat_no;
};


/// \brief per branch least-squares ('BLS') root
///
struct prob_gtor_root_lsprob : prob_gtor<prob_t> {

  typedef prob_gtor_root_lsprob self_t;
  typedef prob_gtor<prob_t> base_t;

  prob_gtor_root_lsprob(const root_gtor_sml_share& smlm,
                        const bool is_leastsq,
                        const unsigned root_cat_no)
    : base_t(smlm.state_size(),0), _smlm(smlm), _is_leastsq(is_leastsq),
      is_assigned_cat(false), assigned_cat_no(0) {

    // convention is all assigned cats are 1:1 mapped to root:
    if(_smlm.get_cat_manager().get_assigned_cat_size()){
      is_assigned_cat=true;
      assigned_cat_no=root_cat_no;
    }
  }

  prob_gtor_root_lsprob(const self_t& s,
                        const root_gtor_sml_share& smlm)
    : base_t(s), _smlm(smlm), _is_leastsq(s._is_leastsq),
      is_assigned_cat(s.is_assigned_cat), assigned_cat_no(s.assigned_cat_no) {}

  virtual self_t* clone(const root_gtor_sml_share& smlm) const { return new self_t(*this,smlm); }

private:
  virtual void get_param_from_distro() const {}

  virtual void param_update(const bool is_secondary_update){
    if(is_secondary_update) _smlm.get_root_node_prob(_distro,_is_leastsq,is_assigned_cat,assigned_cat_no);
  }

  const root_gtor_sml_share _smlm;
  const bool _is_leastsq;
  bool is_assigned_cat;
  unsigned assigned_cat_no;
};



#ifdef BRANCH_SPECIFIC_OBS
struct prob_gtor_root_lsprob : prob_gtor<prob_t> {

  typedef prob_gtor_root_lsprob self_t;
  typedef prob_gtor<prob_t> base_t;

  prob_gtor_root_lsprob(const root_gtor_sml_share& smlm,
                        const bool is_leastsq)
    : base_t(smlm.state_size(),0), _smlm(smlm), _is_leastsq(is_leastsq),
      is_assigned_cat(false), assigned_cat_no(0), _node_state_prob(smlm.branch_size()+1,smlm.state_size()) {}

  prob_gtor_root_lsprob(const self_t& s,
                        const root_gtor_sml_share& smlm)
    : base_t(s), _smlm(smlm), _is_leastsq(s._is_leastsq),
      is_assigned_cat(s.is_assigned_cat), assigned_cat_no(s.assigned_cat_no),
      _node_state_prob(s._node_state_prob) {}

  virtual self_t* clone(const root_gtor_sml_share& smlm) const { return new self_t(*this,smlm); }

  const prob_t* const * node_state_prob() const {
    return _node_state_prob.val;
  }

private:
  virtual void get_param_from_distro() const {}

  virtual void param_update(){
    _smlm.get_root_node_prob_nspround(_distro,_is_leastsq,is_assigned_cat,assigned_cat_no,_node_state_prob.val);
  }

  const root_gtor_sml_share _smlm;
  const bool _is_leastsq;
  bool is_assigned_cat;
  unsigned assigned_cat_no;
  simple_matrix<prob_t> _node_state_prob;
};
#endif

#endif
