// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: rate_gtor.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "rate_gtor.h"
#include "subs_ml_ptol.h"
#include "substk_exception.h"
#include "util/general/die.h"
#include "util/general/log.h"
#include "util/math/matrix_util_io.h"
#include "util/math/random_util.h"
#include "util/math/test_float.h"

#include <iomanip>
#include <ostream>
#include <sstream>
#include <string>




unsigned
rate_gtor_base::
param_size(const PARAM_VIEW::index_t pv) const {

  if       (pv == PARAM_VIEW::ALL){
    return _psize;

  } else if(pv == PARAM_VIEW::MIN_PARAM ||
            pv == PARAM_VIEW::TRAINABLE){
    unsigned pst(0);
    unsigned ps(param_size());
    for(unsigned i(0);i<ps;++i) if(_is_train_param[i]) pst++;
    if(pv == PARAM_VIEW::TRAINABLE) { return pst; }
    else                            { return pst+pseudo_param_size(); }

  } else if(pv == PARAM_VIEW::INDY_MIN_PARAM){
    unsigned pst(0);
    bool is_in_block(false);
    bool is_first_in_block(false);
    const unsigned ps(param_size());
    for(unsigned i(0);i<ps;++i) {
      if(_is_free_block_start[i]) {
        is_in_block=true;
        is_first_in_block=true;
      }

      if(_is_train_param[i]){
        if(is_in_block && is_first_in_block){
          is_first_in_block=false;
        } else { pst++; }
      }

      if(_is_free_block_stop[i]) is_in_block=false;
    }
    return pst;

  } else { die("unknown param view"); }
}



/// \brief debugging pretty print of rates:
///
void
rate_gtor_base::
dump(const char* label,
     std::ostream& os) const {

  os << "rate parameters: " << label << "\n";
  for(unsigned i(0);i<param_size();++i){
    os << "[ " << i << " ] : " << _param[i] << "\n";
  }
  os << "\n";

  const unsigned ss(state_size());

  smlfloat* r(new smlfloat[ss*ss]);
  rates(r);

  static const int precision(4);
  matrix_dump(r,ss,os,precision);

  delete [] r;
}


void
rate_gtor_base::
check_param() const {
  for(unsigned i(0);i<param_size();++i) {
    if(is_float_invalid(_param[i])){
      std::ostringstream oss;
      oss << "Invalid model parameter: " << i << " " << _param[i] ;
      log_os << oss.str() << "\n";
      throw substk_exception(oss.str().c_str());
    }
  }
}


void
rate_gtor_base::
reset_param(const DENOVO_MODEL::index_t dm,
            const bool is_protected_call){
  // randomization parameters
  //
  static const smlfloat rate_min(0.01);
  static const smlfloat rate_max(0.66);

  // derived parameters:
  //
  static const smlfloat rate_range(rate_max-rate_min);

  if(dm == DENOVO_MODEL::RANDOM){
    const unsigned ps(param_size());
    for(unsigned i(0);i<ps;++i){
      _param[i] = random_uniform()*rate_range+rate_min;
    }

  } else if(dm == DENOVO_MODEL::START){
    set_uniform_param_state(1.);

  } else { die("unknown denovo model"); }

  if(! is_protected_call) {
    fix_input_param();
  }
}



void
rate_gtor_base::
get_typed_cat_dependent_prior_prob(const SITE_CAT_TYPE::index_t sct,
                                   prob_t* p_prob) const {

  const unsigned n_typed_cats(get_cat_manager().get_typed_cat_size(sct));
  for(unsigned tc(0);tc<n_typed_cats;++tc) p_prob[tc] = 0.;

  const unsigned n_site_cats(category_size_site());
  std::vector<prob_t> site_cat_prior_pdistro(n_site_cats);

  get_site_category_pdistro(site_cat_prior_pdistro.begin());

  for(unsigned sc(0);sc<n_site_cats;++sc){
    const unsigned typed_cat(get_typed_cat_from_site_cat(sct,sc));
    p_prob[typed_cat] += site_cat_prior_pdistro[sc];
  }

  pdistro_norm(p_prob,p_prob+n_typed_cats);
}


void
rate_gtor_base::
set_pseudo_param(const unsigned param_no,
                 const smlfloat /*val*/){
  if(param_no>=pseudo_param_size()){
    die("Bad pseudoparam number");
  }
}




void
rate_gtor_base::
set_param_size(const unsigned p){
  const unsigned old_psize(param_size());
  smlfloat* old_param = _param;

  _psize = p;
  _param = new smlfloat[_psize];

  _is_train_param.resize(_psize,true);
  _is_free_block_start.resize(_psize,false);
  _is_free_block_stop.resize(_psize,false);

  for(unsigned i(0);i<std::min(old_psize,_psize);++i){
    _param[i] = old_param[i];
  }
  for(unsigned i(old_psize);i<_psize;++i) { _param[i] = 0.; }

  if(old_param) { delete [] old_param; }
}
