// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: root_gtor.h 741 2007-08-14 02:39:19Z ctsa $

/// \file

#ifndef __ROOT_GTOR_H
#define __ROOT_GTOR_H

#include "cat_manager.h"
#include "denovo_types.h"
#include "prob_gtor.h"
#include "rate_gtor_options.h"
#include "subs_ml_model_comex.h"
#include "subs_ml_types.h"

#include <cassert>

#include <iosfwd>


namespace ROOT_GTOR_MODEL {
  enum index_t { FULL,
                 CODON,
                 NUC,
                 CODON_DINUC,
                 NUC_POS,
                 PPROB,
                 OBS_TIME_AVG,
                 LSPROB};
}


/// \brief subs_ml_model root distribution handler
///
struct root_gtor {

  typedef root_gtor self_t;
  typedef prob_gtor<smlfloat> ptor_t;

  root_gtor(std::istream& is,
            const root_gtor_sml_share& smls)
    : _smls(smls),
      _rogm(ROOT_GTOR_MODEL::FULL),
      _ptor(0) {
    load_state(is);
  }

  root_gtor(const ROOT_GTOR_MODEL::index_t rogm,
            const root_gtor_sml_share& smls)
    : _smls(smls), _rogm(rogm), _ptor(0) {
    init();
  }

  root_gtor(const self_t& s,
            const root_gtor_sml_share& smls)
    : _smls(smls), _rogm(s._rogm),
      _ptor(new ptor_t*[root_cat_size()]) {
    for(unsigned i(0);i<root_cat_size();++i){
      _ptor[i]=s._ptor[i]->clone(smls);
    }
  }

  ~root_gtor() { clear(); }

private:
  root_gtor(const self_t& s);
  self_t& operator=(const self_t& rhs);
public:

  unsigned
  state_size() const {
    return _smls.state_size();
  }

  unsigned
  param_size(const PARAM_VIEW::index_t pv) const {
    unsigned pst(0);
    for(unsigned i(0);i<root_cat_size();++i){
      const ptor_t& p(root_cat_ptor(i));
      pst += p.param_size(pv);
    }
    return pst;
  }

  void
  lock_root() {
    for(unsigned i(0);i<root_cat_size();++i){
      ptor_t& p(root_cat_ptor(i));
      p.lock_root();
    }
  }

  template <typename OutputIterator>
  OutputIterator
  param_state(OutputIterator x,
              const PARAM_VIEW::index_t pv) const {
    for(unsigned i(0);i<root_cat_size();++i){
      const ptor_t& p(root_cat_ptor(i));
      x = p.param_state(x,pv);
    }
    return x;
  }

  template <typename InputIterator>
  InputIterator
  set_param_state(InputIterator x,
                  const PARAM_VIEW::index_t pv,
                  const bool is_secondary_update = true) {
    for(unsigned i(0);i<root_cat_size();++i){
      ptor_t& p(root_cat_ptor(i));
      x = p.set_param_state(x,pv,is_secondary_update);
    }
    return x;
  }

  template <typename OutputIterator>
  OutputIterator
  is_train_param(OutputIterator x) const {
    for(unsigned i(0);i<root_cat_size();++i){
      const ptor_t& p(root_cat_ptor(i));
      x = p.is_train_param(x);
    }
    return x;
  }

  template <typename InputIterator>
  InputIterator
  set_is_train_param(InputIterator x) {
    for(unsigned i(0);i<root_cat_size();++i){
      ptor_t& p(root_cat_ptor(i));
      x = p.set_is_train_param(x);
    }
    return x;
  }

  // set all root_cats to the same distro:
  // !intentionally adjusts all params, including no-train!
  template <typename ForwardIterator>
  void set_distro(ForwardIterator x) {
    for(unsigned i(0);i<root_cat_size();++i){
      ptor_t& p(root_cat_ptor(i));
      p.set_distro(x);
    }
  }

  // !intentionally adjusts all params, including no-train!
  template <typename InputIterator>
  void set_all_root_cat_distro(InputIterator x) {
    for(unsigned i(0);i<root_cat_size();++i){
      ptor_t& p(root_cat_ptor(i));
      x = p.set_distro(x);
    }
  }


  template <typename InputIterator>
  InputIterator
  attach_ci(InputIterator ci) {
    for(unsigned i(0);i<root_cat_size();++i){
      ptor_t& p(root_cat_ptor(i));
      ci = p.attach_ci(ci);
    }
    return ci;
  }

  smlfloat param_variance(unsigned i) const;

  unsigned root_cat_size() const { return _smls.get_cat_manager().get_typed_cat_size(SITE_CAT_TYPE::ROOT); }

  const prob_t*
  root_cat_distro(const unsigned root_cat) const {
    return root_cat_ptor(root_cat).distro();
  }

  const prob_t*
  site_cat_distro(const unsigned site_cat) const {
    const unsigned root_cat(_smls.get_cat_manager().get_typed_cat_no(SITE_CAT_TYPE::ROOT,site_cat));
    return root_cat_distro(root_cat);
  }

#ifdef BRANCH_SPECIFIC_OBS
  const prob_t* const *
  site_cat_node_state_prob(const unsigned site_cat) const {
    const unsigned root_cat(_smls.get_cat_manager().get_typed_cat_no(SITE_CAT_TYPE::ROOT,site_cat));
    return root_cat_node_state_prob(root_cat);
  }
#endif

  void reset_param(const DENOVO_MODEL::index_t dm);

  void store_state(std::ostream& os) const;

private:
  void load_state(std::istream& is);

#ifdef BRANCH_SPECIFIC_OBS
  const prob_t* const *
  root_cat_node_state_prob(const unsigned root_cat) const;
#endif

  ptor_t&
  root_cat_ptor(const unsigned root_cat) {
    assert(root_cat<root_cat_size());
    return *_ptor[root_cat];
  }

  const ptor_t&
  root_cat_ptor(const unsigned root_cat) const {
    assert(root_cat<root_cat_size());
    return *_ptor[root_cat];
  }

  void
  clear() {
    if(_ptor){
      for(unsigned i(0);i<root_cat_size();++i){ delete _ptor[i]; }
      delete [] _ptor;
      _ptor = 0;
    }
  }

  void
  init() {
    _ptor = new ptor_t*[root_cat_size()];

    for(unsigned i(0);i<root_cat_size();++i){
      _ptor[i]=prob_gtor_factory(i);
    }
  }

  ptor_t* prob_gtor_factory(const unsigned root_cat_no) const;

  const root_gtor_sml_share _smls;
  ROOT_GTOR_MODEL::index_t _rogm;
  prob_gtor<smlfloat>** _ptor;
};


#endif
