// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: site_data_fastlup_util.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "bi_tree_type.h"
#include "cat_info.h"
#include "catmap_util.h"
#include "lhood_site_cached_optorder.h"
#include "site_data.h"
#include "site_data_fastlup_util.h"

#include <algorithm>
#include <iostream>
#include <set>


// transfer info from map to fixed array
//
static
void
site_data_to_site_data_fastlup(const bi_tree_sml_type& tree,
                               const site_data& sd,
                               const unsigned n_states,
                               const unsigned n_model_site_cats,
                               const std::vector<unsigned>& catlup,
                               site_data_fastlup& sdf){

  const unsigned n_seqs(tree.n_leaves());

  if(sd.taxid.size() < n_seqs){
    std::cerr << "ERROR:: fewer seqs in data ("<< sd.taxid.size() << ") than in tree (" << n_seqs << ")\n";
    abort();
  }


  const unsigned n_sites(sd.site_count.size());
  sdf.init(n_sites,n_seqs,n_model_site_cats);

  // used to track which skipped data categories we've already warned on:
  std::set<unsigned> is_cat_warning;

  std::vector<unsigned> sd_org_tree_order(n_seqs);
  for(unsigned j(0);j<n_seqs;++j){
    sd_org_tree_order[j] = sd.taxid.getid(tree.leaf[j]->label);
  }

  unsigned n(0);
  site_count_map::const_iterator i,i_end(sd.site_count.end());
  for(i=sd.site_count.begin();i!=i_end;++i,++n){
    const site_code& sc(i->first);
    const site_count_data_type& site_count_data(i->second);

    // convenient ref to current sdf cell:
    site_data_fastlup_cell& dn(sdf.data[n]);

    // initialize sdf cell:
    dn.unassigned_cat_count = 0;
    for(unsigned c(0);c<n_model_site_cats;++c) dn.assigned_cat_count[c] = 0;

    // transfer site type to current sdf cell:
    for(unsigned j(0);j<n_seqs;++j){
      dn.index[j] = sc.get_taxid(sd_org_tree_order[j]);
    }

    // iterate through all cat counts for this site type:
    site_count_data_type::const_iterator j,j_end(site_count_data.end());
    for(j=site_count_data.begin();j!=j_end;++j){
      const unsigned& data_cat_no(j->first.cat_no);
      const unsigned& count(j->second);

      const unsigned model_cat_no(catlup[data_cat_no]);

      if(model_cat_no == UNASSIGNED_CAT_ID) {
        dn.unassigned_cat_count += count;
      } else if(model_cat_no <= n_model_site_cats) {
        assert(model_cat_no>0);
        dn.assigned_cat_count[model_cat_no-1] += count;
      } else {
        if(is_cat_warning.find(data_cat_no) == is_cat_warning.end()){
          std::cerr << "WARNING:: discarding unmapped data category# " << data_cat_no << "\n";
          is_cat_warning.insert(data_cat_no);
        }
      }
    }
  }

  // find optimal sdf org order for cached lhood calculation:
  lhood_site_cached_optorder(tree,n_states,sdf);

  sdf.fix();


  // create left and right side pair count maps in sorted sdf,
  // assume cats do not exist:

  // first get a site lookup map, sdfl
  typedef  std::map<site_code,unsigned> qmap;
  qmap sdfl;

  {
    site_code sc(n_seqs);

    for(unsigned j(0);j<n_sites;++j){
      for(unsigned s(0);s<n_seqs;++s){
        sc.set_taxid(sd_org_tree_order[s],sdf.data_core[j].index[s]);
      }
      sdfl[sc] = j;
    }
  }

  // step through and accumulate pair left sides write out
  // pair-info:
  pair_count_map::const_iterator pi,pi_end=sd.pair_count.end();

  const qmap::const_iterator qend(sdfl.end());

  sdf.nb_left.clear();
  sdf.nb_right.clear();

  for(pi=sd.pair_count.begin();pi!=pi_end;++pi){
    const site_pair_type sp(pi->first.first);
    const cat_pair_type cp(pi->first.second);

    // don't use categories (for now):
    if(cp.first  != UNASSIGNED_CAT_ID ||
       cp.second != UNASSIGNED_CAT_ID) continue;

    const unsigned count(pi->second);

    // skip sites not in site_data
    /// \todo fix so that edge site types can get post prob calculated even though
    /// they won't be used in final lhood calc
    ///
    const qmap::const_iterator i1(sdfl.find(sp.first));
    const qmap::const_iterator i2(sdfl.find(sp.second));

    // on second thought, this shouldn't happen!!
    assert(i2 != qend);

    if(i1 == qend || i2 == qend) continue;

    const unsigned s1(i1->second);
    const unsigned s2(i2->second);

    // temporary:
    sdf.nb_left.push_back(neighbor_count(s2,s1,count));
    sdf.nb_right.push_back(neighbor_count(s1,s2,count));
  }

  std::sort(sdf.nb_left.begin(),sdf.nb_left.end());
  std::sort(sdf.nb_right.begin(),sdf.nb_right.end());

#if 0
  for(unsigned j(0);j<n_sites;++j){
    std::cerr << "site: " << j;
    for(unsigned s(0);s<n_seqs;++s){
      std::cerr << " " << sdf.data_core[j].index[s];
    }
    std::cerr << "\n";
  }

  const unsigned nbs(sdf.nb_left.size());
  for(unsigned j(0);j<nbs;++j){
    std::cerr << "nb_left: " << j << " " << sdf.nb_left[j].id1 << " "
              << sdf.nb_left[j].id2 << " " << sdf.nb_left[j].count << "\n";
  }

  const unsigned nbs2(sdf.nb_right.size());
  for(unsigned j(0);j<nbs2;++j){
    std::cerr << "nb_right: " << j << " " << sdf.nb_right[j].id1 << " "
              << sdf.nb_right[j].id2 << " " << sdf.nb_right[j].count << "\n";
  }
#endif
}



void
site_data_to_site_data_fastlup(const bi_tree_sml_type& tree,
                               const site_data& sd,
                               const unsigned n_states,
                               const unsigned n_model_site_cats,
                               const data_catmap_type& catmap,
                               site_data_fastlup& sdf){

  // convert catmap to fast lup array:
  std::vector<unsigned> catlup;
  catmap_to_catlup(sd,n_model_site_cats,catmap,catlup);

  site_data_to_site_data_fastlup(tree,sd,n_states,n_model_site_cats,catlup,sdf);
}



// eliminates cat info in reduced form!
//
void
site_data_fastlup_state_reduction(const bi_tree_sml_type& tree,
                                  const unsigned reduced_n_states,
                                  const site_data_fastlup& sdf_in,
                                  const unsigned* index_translation,
                                  site_data_fastlup& sdf_out){

  const unsigned n_cats(sdf_in.n_cats);
  const unsigned n_seqs(sdf_in.n_orgs);

  site_data sd;

  // insert remapped counts into sd
  for(unsigned i(0);i<sdf_in.len;++i){
    site_code sc(n_seqs);
    for(unsigned t(0);t<n_seqs;++t){
      sc.set_taxid(t,index_translation[sdf_in.data[i].index[t]]);
    }

    {
      const unsigned count(sdf_in.data[i].unassigned_cat_count);
      sd.site_count[sc][count_data_code(UNASSIGNED_CAT_ID)] = count;
    }

    const unsigned asize(n_cats);
    for(unsigned j(0);j<asize;++j){
      const unsigned cat_no(j);
      const unsigned count(sdf_in.data[i].assigned_cat_count[j]);
      sd.site_count[sc][count_data_code(cat_no+1)] += count;
    }
  }

  // fudgin' stuff...

  // pass a catlup that matches the current cat order:
  std::vector<unsigned> catlup(n_cats);
  for(unsigned i(0);i<n_cats;++i) catlup[i] = i;

  // give sd taxid's that match the tree order:
  std::vector<unsigned> sd_org_tree_order(n_seqs);
  for(unsigned j(0);j<n_seqs;++j){
    sd.taxid.assignid(tree.leaf[j]->label);
  }

  site_data_to_site_data_fastlup(tree,sd,reduced_n_states,n_cats,catlup,sdf_out);
}
