// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: site_data_report.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "site_data.h"
#include "site_data_report.h"
#include "site_data_stats.h"
#include "stationary_pdistro.h"
#include "subs_ml_model.h"
#include "util/bio/bioseq_util.h"
#include "util/bio/bioseq_util_pdf_conversions.h"

#include <ostream>
#include <sstream>

using namespace std;

#if 0
static
void
report_nscodon_distro(const prob_t nsp[],
                      const char* label,
                      std::ostream& os){

  os << label << "\n";
  os << "codon distro:\n";

  NSCODON::print nscodon_syms[NSCODON::SIZE];
  for(unsigned i(0);i<NSCODON::SIZE;++i){ nscodon_syms[i] = NSCODON::print(static_cast<NSCODON::index_t>(i)); }

  pdistro_report(nsp,NSCODON::SIZE,nscodon_syms,os);
}
#endif


static
void
dump_seqs_nscodon(const site_data& sd,
                  std::ostream& os){

  const unsigned n_taxa(sd.n_taxa());

  for(unsigned t(0);t<n_taxa;++t) {
    os << sd.taxid.getstr(t) << " ";
  }
  os << "\n";

  site_count_map::const_iterator i,i_end;
  i=sd.site_count.begin();
  i_end=sd.site_count.end();

  for(;i!=i_end;++i){
    site_count_data_type::const_iterator j,j_end;
    j=i->second.begin();
    j_end=i->second.end();

    unsigned count(0);
    for(;j!=j_end;++j){ count += j->second; }

    for(unsigned t(0);t<n_taxa;++t){
      const NSCODON::index_t c(static_cast<NSCODON::index_t>(i->first.get_taxid(t)));
      os << NSCODON::print(c,true) << " ";
    }

    os << count << "\n";
  }
}



static
void
dump_seqs_nsc4(const site_data& sd,
               std::ostream& os){

  const unsigned n_taxa(sd.n_taxa());

  for(unsigned t(0);t<n_taxa;++t){
    os << sd.taxid.getstr(t) << " ";
  }
  os << "\n";

  site_count_map::const_iterator i,i_end;
  i=sd.site_count.begin();
  i_end=sd.site_count.end();

  for(;i!=i_end;++i){
    site_count_data_type::const_iterator j,j_end;
    j=i->second.begin();
    j_end=i->second.end();

    unsigned count(0);
    for(;j!=j_end;++j){ count += j->second; }

    for(unsigned t(0);t<n_taxa;++t){
      const NSC4::index_t f(static_cast<NSC4::index_t>(i->first.get_taxid(t)));
      const NSCODON::index_t c(NSC4::decode_nscodon(f));
      const NUC::index_t n(NSC4::decode_nuc(f));

      if(sd.sm == SITE_MODEL::NSC4PRE) os << NUC::syms[n] << "+";
      os << NSCODON::print(c);
      if(sd.sm == SITE_MODEL::NSC4POST) os << "+" << NUC::syms[n];
      os <<  " ";
    }

    os << count << "\n";
  }
}



template <typename SYM_TYPE>
void
report_org_distros(const unsigned n_orgs,
                   const unsigned n_states,
                   const prob_t* pdf, // [n_org*n_states]
                   const name_id_lup& labels,
                   const SYM_TYPE* syms,
                   std::ostream& os){

  for(unsigned i(0); i<n_orgs;++i){
    if(i!=0) os << " ";
    os << labels.getstr(i);
  }
  os << "\n";

  os << std::setprecision(5) << std::left << fixed;
  for(unsigned i(0);i<n_states;++i){
    os << syms[i] << " : ";
    for(unsigned j(0);j<n_orgs;++j){
      os << std::setw(8) << pdf[i+j*n_states] << " ";
    }
    os << "\n";
  }
  os << "\n";
}


static
void
site_data_synonymous_site_count(const site_data& sd,
                                bool& is_coding,
                                unsigned& synonymous_site_count){

  synonymous_site_count = 0;
  is_coding = SITE_MODEL::is_coding(sd.sm);
  if(! is_coding) return;

  const unsigned n_seqs(sd.n_taxa());

  site_count_map::const_iterator i,i_end=sd.site_count.end();

  for(i=sd.site_count.begin();i!=i_end;++i){
    site_count_data_type::const_iterator j,j_end=i->second.end();

    // check if this is a synon site:
    bool is_synon(true);
    {
      NSAA::index_t last_aa(NSAA::X);
      for(unsigned t(0);t<n_seqs;++t){
        const unsigned state(i->first.get_taxid(t));
        const NSCODON::index_t nsc(decode_nscodon(sd.sm,state));
        const NSAA::index_t aa(codon_trans_known(nsc));
        if(t!=0 && aa!=last_aa) {
          is_synon=false;
          break;
        }
        last_aa=aa;
      }
    }

    // sum over all categories:
    if(is_synon){
      for(j=i->second.begin();j!=j_end;++j) synonymous_site_count += j->second;
    }
  }
}


static
void
site_data_write_counts(const site_data& sd,
                       std::ostream& os){

  unsigned site_count(0);
  unsigned conserved_site_count(0);
  unsigned unique_site_count(0);
  site_data_counts(sd,site_count,conserved_site_count,unique_site_count);

  os << "total sites: " << site_count << "\n";
  os << "conserved sites: " << conserved_site_count << "\n";
  os << "unique sites: " << unique_site_count << "\n";

  bool is_coding(false);
  unsigned synonymous_site_count(0);
  site_data_synonymous_site_count(sd,is_coding,synonymous_site_count);
  if(is_coding) os << "synonymous sites: " << synonymous_site_count << "\n";
  os << "\n";

  const unsigned n_cats(sd.max_cat()+1);
  std::vector<unsigned> data_cat_count(n_cats,0);
  site_data_2_cat_count(sd,n_cats,data_cat_count);
  os << "category total sites:\n";
  for(unsigned c(0);c<n_cats;++c) {
    os << c << " " << sd.id_2_catlabel.find(c)->second << " " << data_cat_count[c] << "\n";
  }
  os << "\n\n";
}




static
void
report_org_nscodon_pdf(const unsigned n_org,
                       const prob_t* org_nscodon_pdf,
                       const name_id_lup& labels,
                       std::ostream& os){

  // nucs:
  prob_t* taxa_nuc_pdf = new prob_t[n_org*NUC::SIZE];
  for(unsigned i(0);i<n_org;++i) nscodon_pdf_2_nuc_pdf(org_nscodon_pdf+i*NSCODON::SIZE,taxa_nuc_pdf+i*NUC::SIZE);

  report_org_distros<char>(n_org,NUC::SIZE,taxa_nuc_pdf,labels,NUC::syms,os);

  // nucs by codon position:
  for(unsigned k(0);k<CODON::BASE_SIZE;++k){
    for(unsigned i(0);i<n_org;++i) nscodon_pdf_2_nuc_pdf_pos(org_nscodon_pdf+i*NSCODON::SIZE,k,taxa_nuc_pdf+i*NUC::SIZE);
    os << "CODON POSITION: " << k << "\n";
    report_org_distros<char>(n_org,NUC::SIZE,taxa_nuc_pdf,labels,NUC::syms,os);
  }
  delete [] taxa_nuc_pdf;

  // aas
  prob_t* taxa_nsaa_pdf = new prob_t[n_org*NSAA::SIZE];
  for(unsigned i(0);i<n_org;++i) nscodon_pdf_2_nsaa_pdf(org_nscodon_pdf+i*NSCODON::SIZE,taxa_nsaa_pdf+i*NSAA::SIZE);

  report_org_distros<char>(n_org,NSAA::SIZE,taxa_nsaa_pdf,labels,NSAA::syms,os);
  delete [] taxa_nsaa_pdf;

  // codons
  NSCODON::print nscodon_syms[NSCODON::SIZE];
  for(unsigned i(0);i<NSCODON::SIZE;++i){ nscodon_syms[i] = NSCODON::print(static_cast<NSCODON::index_t>(i),true); }

  report_org_distros<NSCODON::print>(n_org,NSCODON::SIZE,org_nscodon_pdf,labels,nscodon_syms,os);
}




static
void
analyze_aa_from_nscodon_pdf(const smlfloat* nscodon_pdf,
                            std::ostream& os){

  static const smlfloat nsexp(1./static_cast<smlfloat>(NSCODON::SIZE));
  static bool is_init(false);
  static map<NSAA::index_t,unsigned> nsaa_codon_count;

  // get codons per aa:
  if(! is_init){
    for(unsigned i(0);i<NSCODON::SIZE;++i){
      NSCODON::index_t ni = static_cast<NSCODON::index_t>(i);
      nsaa_codon_count[codon_trans_known(ni)]++;
    }
    is_init=true;
  }

  // get aa distro
  smlfloat nsaa_pdf[NSAA::SIZE];
  nscodon_pdf_2_nsaa_pdf(nscodon_pdf,nsaa_pdf);

  // get neutral expected aa distro from nuc bg
  smlfloat nuc_pdf[NSAA::SIZE];
  nscodon_pdf_2_nuc_pdf(nscodon_pdf,nuc_pdf);

  smlfloat nscodon_expect_pdf[NSCODON::SIZE];
  nuc_pdf_2_nscodon_pdf(nuc_pdf,nscodon_expect_pdf);

  smlfloat nsaa_expect_pdf[NSCODON::SIZE];
  nscodon_pdf_2_nsaa_pdf(nscodon_expect_pdf,nsaa_expect_pdf);


  os << "AA    OBS    EXP    ln2(O/E)    EXP_NUC_BG    ln2(O/E_N)\n";
  for(unsigned i(0);i<NSAA::SIZE;++i){
    // get naive (uniform nuc distro) neutral aa expectation:
    const smlfloat nsaa_exp = nsexp*nsaa_codon_count[static_cast<NSAA::index_t>(i)];
    os << NSAA::syms[i] << " : " << nsaa_pdf[i] << "   ";

    os << nsaa_exp << " ";
    os.setf(ios::showpos);
    os << std::log(nsaa_pdf[i]/nsaa_exp)/M_LN2 << "   ";
    os.unsetf(ios::showpos);

    os << nsaa_expect_pdf[i] << " ";
    os.setf(ios::showpos);
    os << std::log(nsaa_pdf[i]/nsaa_expect_pdf[i])/M_LN2 << "\n";
    os.unsetf(ios::showpos);
  }
  os << "\n";
}



static
void
analyze_nscodon_pdf(const smlfloat* nscodon_pdf,
                    std::ostream& os){

  static const smlfloat nsexp = 1./static_cast<smlfloat>(NSCODON::SIZE);

  smlfloat nuc_pdf[NSAA::SIZE];
  nscodon_pdf_2_nuc_pdf(nscodon_pdf,nuc_pdf);

  smlfloat nscodon_expect_pdf[NSCODON::SIZE];
  nuc_pdf_2_nscodon_pdf(nuc_pdf,nscodon_expect_pdf);

  os << "CODON    OBS      EXP    ln2(O/E)      EXP_NUC_BG    ln2(O/E_N)\n";
  for(unsigned i(0);i<NSCODON::SIZE;++i){
    NSCODON::index_t ni = static_cast<NSCODON::index_t>(i);
    os << NSCODON::print(ni,true) << " : " << nscodon_pdf[i] << "   "
       << nsexp << " ";
    os.setf(ios::showpos);
    os << std::log(nscodon_pdf[i]/nsexp)/M_LN2 << "   ";
    os.unsetf(ios::showpos);
    os << nscodon_expect_pdf[i] << " ";
    os.setf(ios::showpos);
    os << std::log(nscodon_pdf[i]/nscodon_expect_pdf[i])/M_LN2 << "\n";
    os.unsetf(ios::showpos);
  }
  os << "\n";
}



void
site_data_report(const site_data& sd,
                 const bool is_pretty_print_data,
                 std::ostream& os){

  const SITE_MODEL::index_t sm(sd.sm);

  os << "site-type: " << SITE_MODEL::label(sm) << "\n";
  os << "\n";

  site_data_write_counts(sd,os);

  const bool is_c4_site(sd.sm == SITE_MODEL::NSC4PRE ||
                        sd.sm == SITE_MODEL::NSC4POST);

  const unsigned n_org(sd.n_taxa());

  if(sd.sm == SITE_MODEL::NUC){
    prob_t* nuc_pdf = new prob_t[n_org*NUC::SIZE];
    site_data_2_taxa_state_pdf(sd,NUC::SIZE,nuc_pdf);
    report_org_distros<char>(n_org,NUC::SIZE,nuc_pdf,sd.taxid,NUC::syms,os);
    delete [] nuc_pdf;
    return;
  }

  if(sd.sm == SITE_MODEL::DINUC){

    DINUC::print dinuc_syms[DINUC::SIZE];
    for(unsigned i(0);i<DINUC::SIZE;++i){ dinuc_syms[i] = DINUC::print(i); }
    {
      prob_t* dinuc_pdf = new prob_t[n_org*DINUC::SIZE];
      site_data_2_taxa_state_pdf(sd,DINUC::SIZE,dinuc_pdf);
      report_org_distros<DINUC::print>(n_org,DINUC::SIZE,dinuc_pdf,sd.taxid,dinuc_syms,os);
      delete [] dinuc_pdf;
    }
    { // more detailed combo and margin reports:
      prob_t all_dinuc_pdf[DINUC::SIZE];
      site_data_2_state_pdf(sd,DINUC::SIZE,all_dinuc_pdf);

      os << "Combined taxa pdf:\n";
      pdistro_report(all_dinuc_pdf,DINUC::SIZE,dinuc_syms,os);

      prob_t all_dinuc_cond_pdf[DINUC::SIZE];
      dinuc_pdf_2_conditioned_dinuc_pdf(all_dinuc_pdf,0,all_dinuc_cond_pdf);

      prob_t n0_stationary_pdistro[NUC::SIZE];
      get_stationary_pdistro_from_pmatrix(n0_stationary_pdistro,all_dinuc_cond_pdf,NUC::SIZE);

      os << "nx+n0 : p(n0|nx):\n";
      pdistro_report(all_dinuc_cond_pdf,DINUC::SIZE,dinuc_syms,os);

      dinuc_pdf_2_conditioned_dinuc_pdf(all_dinuc_pdf,1,all_dinuc_cond_pdf);
      os << "nx+n0 : p(nx|n0):nuc \n";
      pdistro_report(all_dinuc_cond_pdf,DINUC::SIZE,dinuc_syms,os);


      prob_t all_nuc_pdf[NUC::SIZE];
      dinuc_pdf_2_nuc_pdf_pos1(all_dinuc_pdf,all_nuc_pdf);
      os << "n0 pdf:\n";
      pdistro_report(all_nuc_pdf,NUC::SIZE,NUC::syms,os);

      os << "n0 stationary pdf:\n";
      pdistro_report(n0_stationary_pdistro,NUC::SIZE,NUC::syms,os);
    }
    return;
  }

  {
    prob_t* nscodon_pdf = new prob_t[n_org*NSCODON::SIZE];

    if       (sd.sm == SITE_MODEL::NSCODON){
      site_data_2_taxa_state_pdf(sd,NSCODON::SIZE,nscodon_pdf);
    } else if(is_c4_site){
      prob_t* nsc4_pdf = new prob_t[n_org*NSC4::SIZE];
      site_data_2_taxa_state_pdf(sd,NSC4::SIZE,nsc4_pdf);

      for(unsigned t(0);t<n_org;++t) nsc4_pdf_2_nscodon_pdf(nsc4_pdf+t*NSC4::SIZE,nscodon_pdf+t*NSCODON::SIZE);

      delete [] nsc4_pdf;
    }

    report_org_nscodon_pdf(n_org,nscodon_pdf,sd.taxid,os);

    delete [] nscodon_pdf;

    if(is_c4_site){
      simple_array<prob_t> nsc4_pdf(n_org*NSC4::SIZE);
      site_data_2_taxa_state_pdf(sd,NSC4::SIZE,nsc4_pdf.val);

      SITE_MODEL::print nsc4_syms[NSC4::SIZE];
      for(unsigned i(0);i<NSC4::SIZE;++i){ nsc4_syms[i] = SITE_MODEL::print(sd.sm,static_cast<NSC4::index_t>(i)); }

      report_org_distros<SITE_MODEL::print>(n_org,NSC4::SIZE,nsc4_pdf.val,sd.taxid,nsc4_syms,os);
    }
  }

  { // report combined-taxa summary:

    smlfloat all_nscodon_pdf[NSCODON::SIZE];

    if       (sd.sm == SITE_MODEL::NSCODON){
      site_data_2_state_pdf(sd,NSCODON::SIZE,all_nscodon_pdf);
    } else if(is_c4_site){
      smlfloat all_nsc4_pdf[NSC4::SIZE];
      site_data_2_state_pdf(sd,NSC4::SIZE,all_nsc4_pdf);

      nsc4_pdf_2_nscodon_pdf(all_nsc4_pdf,all_nscodon_pdf);
#if 1
      // supplemental ultrabug analysis
      prob_t nx_pdf[NUC::SIZE];
      prob_t n1_pdf[NUC::SIZE];
      prob_t n2_pdf[NUC::SIZE];
      prob_t nxn2_pdf[NUC::SIZE*NUC::SIZE];

      prob_t n1n2_pdf[NUC::SIZE*NUC::SIZE];
      for(unsigned i(0);i<NUC::SIZE;++i){
        nx_pdf[i] = 0.;
        n1_pdf[i] = 0.;
        n2_pdf[i] = 0.;
      }
      for(unsigned i(0);i<(NUC::SIZE*NUC::SIZE);++i){
        nxn2_pdf[i] = 0.;
        n1n2_pdf[i] = 0.;
      }

      for(unsigned i(0);i<NSC4::SIZE;++i){
        NUC::index_t nx;
        NSCODON::index_t c;
        NSC4::decode(nx,c,i);

        NUC::index_t n[3];
        NSCODON::decode(n,c);
        nx_pdf[nx] += all_nsc4_pdf[i];
        n1_pdf[n[1]] += all_nsc4_pdf[i];
        n2_pdf[n[2]] += all_nsc4_pdf[i];
        nxn2_pdf[n[2]+NUC::SIZE*nx] += all_nsc4_pdf[i];
        n1n2_pdf[n[2]+NUC::SIZE*n[1]] += all_nsc4_pdf[i];
      }

      os << "c4 nx distro:\n";
      pdistro_report(nx_pdf,NUC::SIZE,NUC::syms,os);

      os << "c4 distros:\n"
         << "nx-c: obs exp:\n";
      for(unsigned i(0);i<NUC::SIZE;++i){
        for(unsigned j(0);j<NSCODON::SIZE;++j){
          const NSC4::index_t ni(NSC4::encode(static_cast<NUC::index_t>(i),
                                              static_cast<NSCODON::index_t>(j)));
          const smlfloat obs(all_nsc4_pdf[ni]);
          const smlfloat exp(nx_pdf[i]*all_nscodon_pdf[j]);
          os << SITE_MODEL::print(sd.sm,ni) << ": "
             << obs << " "
             << exp << " "
             << std::log(obs/exp) << "\n";
        }
      }
      os << "\n";


      os << "c4 n1 distro:\n";
      pdistro_report(n1_pdf,NUC::SIZE,NUC::syms,os);

      os << "c4 n2 distro:\n";
      pdistro_report(n2_pdf,NUC::SIZE,NUC::syms,os);

      {
        DINUC::print dinuc_syms[DINUC::SIZE];
        for(unsigned i(0);i<DINUC::SIZE;++i){ dinuc_syms[i] = DINUC::print(i); }

        prob_t all_dinuc_cond_pdf[DINUC::SIZE];
        dinuc_pdf_2_conditioned_dinuc_pdf(n1n2_pdf,0,all_dinuc_cond_pdf);

        os << "c4: n1+n2 : p(n2|n1):\n";
        pdistro_report(all_dinuc_cond_pdf,DINUC::SIZE,dinuc_syms,os);

        dinuc_pdf_2_conditioned_dinuc_pdf(n1n2_pdf,1,all_dinuc_cond_pdf);

        os << "c4: n1+n2 : p(n1|n2):\n";
        pdistro_report(all_dinuc_cond_pdf,DINUC::SIZE,dinuc_syms,os);
      }
#if 0
      os << "\n"
         << "double nuc distros:\n"
         << "N-M: obs exp:\n";
      for(unsigned i(0);i<NUC::SIZE;++i){
        for(unsigned j(0);j<NUC::SIZE;++j){
          smlfloat obs=nxn2_pdf[j+NUC::SIZE*i];
          smlfloat exp=nx_pdf[i]*n2_pdf[j];
          os << NUC::syms[i] << "-" << NUC::syms[j] << ": "
             << obs << " "
             << exp << " "
             << std::log(obs/exp) << "\n";
        }
      }
#endif
#endif
    }

    analyze_aa_from_nscodon_pdf(all_nscodon_pdf,os);
    analyze_nscodon_pdf(all_nscodon_pdf,os);
  }

  // dump all data in readable format:
  if(is_pretty_print_data) {
    os << "PRETTY_PRINT_SITES:\n";
    if(sd.sm == SITE_MODEL::NSCODON){
      dump_seqs_nscodon(sd,os);
    } else if(is_c4_site){
      dump_seqs_nsc4(sd,os);
    }
  }
}

#if 0
void
report_nscodon_model(const subs_ml_model& mdl,
                     std::ostream& os){

  os << "root distro:\n";

  prob_t stack_nscodon_pdf[NSCODON::SIZE];

  for(unsigned i(0);i<mdl.get_root_gtor().root_cat_size();++i){
    const prob_t* root_pdf(mdl.get_root_gtor().root_cat_distro(i));
    const prob_t* nscodon_pdf(root_pdf);

    const BIO_PDISTRO::index_t bp(BIO_PDISTRO::convert_from_site_model(mdl.rate_gtor().site_model()));
    if(bp != BIO_PDISTRO::NSCODON){
      bio_pdistro_convert(bp,BIO_PDISTRO::NSCODON,root_pdf,stack_nscodon_pdf);
      nscodon_pdf=stack_nscodon_pdf;
    }

    std::string distro_label;
    {
      std::ostringstream oss;
      format_root_header(i,mdl.tree().root->label.c_str(),oss);
      distro_label = oss.str();
    }

    // nuc
    report_nuc_distro_from_nscodon(nscodon_pdf,distro_label.c_str(),os);

    // nuc by codon pos
    for(unsigned k(0);k<CODON::BASE_SIZE;++k){
      report_nuc_distro_from_nscodon_pos(nscodon_pdf,distro_label.c_str(),k,os);
    }

    // aa
    report_nsaa_distro_from_nscodon(nscodon_pdf,distro_label.c_str(),os);

    // codons
    print_site_model_distro(root_pdf,sm,os);
    report_nscodon_distro(nscodon_pdf,distro_label.c_str(),os);
  }

  mdl.report_time(os);
  os << "\n";

  mdl.rate_gtor().report("ALL-BRANCH",os);
}
#endif

#if 0
void
report_multinuc_model(const subs_ml_model& mdl,
                     std::ostream& os){

  os << "root distro:\n";

  const SITE_MODEL::index_t sm(mdl.rate_gtor().site_model());

  for(unsigned i(0);i<mdl.get_root_gtor().root_cat_size();++i){
    const prob_t* root_pdf(mdl.get_root_gtor().root_cat_distro(i));

    format_root_header(i,mdl.tree().root->label.c_str(),os);
    os << "\n";
    print_site_model_distro(root_pdf,sm,os);

    prob_t nuc_pdf[NUC::SIZE];
    const BIO_PDISTRO::index_t bp(BIO_PDISTRO::convert_from_site_model(sm));
    if(bp != BIO_PDISTRO::NUC){
      bio_pdistro_convert(bp,BIO_PDISTRO::NUC,root_pdf,nuc_pdf);

      format_root_header(i,mdl.tree().root->label.c_str(),os);
      os << "\n";
      print_site_model_distro(nuc_pdf,SITE_MODEL::NUC,os);

    }
  }

  mdl.report_time(os);
  os << "\n";

  mdl.rate_gtor().report("ALL-BRANCH",os);
}
#endif
