// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: site_data_stats.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "site_data.h"
#include "site_data_stats.h"
#include "util/bio/bioseq_util.h"
#include "util/math/prob_util.h"

using namespace std;


// get expected symbol distribution from all site data
void
site_data_2_state_pdf(const site_data& sd,
                      const unsigned n_states,
                      prob_t* distro){

  const unsigned n_seqs(sd.n_taxa());

  memset(distro,0,n_states*sizeof(prob_t));

  site_count_map::const_iterator i,i_end=sd.site_count.end();

  for(i=sd.site_count.begin();i!=i_end;++i){
    site_count_data_type::const_iterator j,j_end=i->second.end();

    for(j=i->second.begin();j!=j_end;++j){
      for(unsigned t(0);t<n_seqs;++t){
        const unsigned state(i->first.get_taxid(t));
        if(state>=n_states) continue;
        distro[state] += static_cast<prob_t>(j->second);
      }
    }
  }

  pdistro_norm(distro,distro+n_states);
}


template <typename T>
void
site_data_2_taxa_state_count_internal(const site_data& sd,
                                      const unsigned n_states,
                                      T* count,
                                      const bool is_accumulate = false,
                                      const bool is_single_cat = false,
                                      const unsigned cat_no = 0){

  const unsigned n_seqs(sd.n_taxa());

  if(! is_accumulate) std::fill(count,count+n_states*n_seqs,static_cast<T>(0));

  site_count_map::const_iterator i,i_end=sd.site_count.end();

  for(i=sd.site_count.begin();i!=i_end;++i){
    site_count_data_type::const_iterator j,j_end=i->second.end();
    for(j=i->second.begin();j!=j_end;++j){
      if(is_single_cat){
        const unsigned data_cat_no(static_cast<unsigned>(j->first.cat_no));
        if(data_cat_no != cat_no) continue;
      }
      for(unsigned t(0);t<n_seqs;++t){
        const unsigned state(i->first.get_taxid(t));
        if(state>=n_states) continue;

        count[t*n_states+state] += static_cast<T>(j->second);
      }
    }
  }
}


// get state counts for each org from site data
void
site_data_2_taxa_state_count(const site_data& sd,
                             const unsigned n_states,
                             unsigned* count,
                             const bool is_accumulate,
                             const bool is_single_cat,
                             const unsigned cat_no){

  site_data_2_taxa_state_count_internal(sd,n_states,count,is_accumulate,is_single_cat,cat_no);
}


// get state distribution for each org from site data
void
site_data_2_taxa_state_pdf(const site_data& sd,
                           const unsigned n_states,
                           prob_t* distro){

  const unsigned n_seqs(sd.n_taxa());

  site_data_2_taxa_state_count_internal(sd,n_states,distro);

  for(unsigned t(0);t<n_seqs;++t) {
    pdistro_norm(distro+t*n_states,distro+(t+1)*n_states);
  }
}



void
site_data_2_cat_state_count(const site_data& sd,
                            const unsigned n_cats,
                            const unsigned n_states,
                            std::vector<unsigned>& cat_state_count){

  const unsigned n_seqs(sd.n_taxa());

  cat_state_count.resize(n_cats*n_states);
  std::fill(cat_state_count.begin(),cat_state_count.end(),0);

  site_count_map::const_iterator i,i_end=sd.site_count.end();

  for(i=sd.site_count.begin();i!=i_end;++i){
    site_count_data_type::const_iterator j,j_end=i->second.end();

    for(j=i->second.begin();j!=j_end;++j){
      const unsigned cat_no(static_cast<unsigned>(j->first.cat_no));
      if(cat_no>=n_cats) continue;
      for(unsigned t(0);t<n_seqs;++t){
        const unsigned state(i->first.get_taxid(t));
        if(state>=n_states) continue;
        cat_state_count[cat_no*n_states+state] += j->second;
      }
    }
  }
}



void
site_data_2_cat_count(const site_data& sd,
                      const unsigned max_cat,
                      std::vector<unsigned>& data_cat_count){

  site_count_map::const_iterator i,i_end=sd.site_count.end();

  for(i=sd.site_count.begin();i!=i_end;++i){
    site_count_data_type::const_iterator j,j_end=i->second.end();

    for(j=i->second.begin();j!=j_end;++j){
      const unsigned cat_no(static_cast<unsigned>(j->first.cat_no));
      if(cat_no>=max_cat) continue;
      data_cat_count[cat_no] += j->second;
    }
  }
}



void
site_data_counts(const site_data& sd,
                 unsigned& site_count,
                 unsigned& conserved_site_count,
                 unsigned& unique_site_count){

  site_count=0;
  conserved_site_count=0;
  unique_site_count=sd.site_count.size();

  const unsigned n_seqs(sd.n_taxa());

  site_count_map::const_iterator i,i_end=sd.site_count.end();

  for(i=sd.site_count.begin();i!=i_end;++i){
    site_count_data_type::const_iterator j,j_end=i->second.end();

    // check if this is a conserved site:
    bool is_conserved(true);
    {
      unsigned last_state(0);
      for(unsigned t(0);t<n_seqs;++t){
        const unsigned state(i->first.get_taxid(t));
        if(t!=0 && state!=last_state) {
          is_conserved=false;
          break;
        }
        last_state=state;
      }
    }

    // sum over all categories:
    for(j=i->second.begin();j!=j_end;++j){
      site_count += j->second;
      if(is_conserved) conserved_site_count += j->second;
    }
  }
}

