// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: subs_ml_model.cc 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "cat_info.h"
#include "report_util.h"
#include "subs_ml_model.h"
#include "subs_ml_print_util.h"
#include "util/general/die.h"
#include "util/general/io_util.h"
#include "util/general/metatags.h"
#include "util/math/prob_util_io.h"

#include <iomanip>
#include <iostream>
#include <sstream>
#include <string>  // for load_state
#include <vector>


const unsigned MODEL_FILE_VERSION(5);


// report a set of times
static
void
report_subs_ml_model_time_instance(const smlfloat scale,
                                   const char* time_label,
                                   const subs_ml_model& mdl,
                                   std::ostream& os) {

  std::vector<irv_t<smlfloat> > branch_time(mdl.get_time_gtor().branch_size());
  for(unsigned i(0);i<mdl.get_time_gtor().branch_size();++i){
    branch_time[i] = scale*mdl.get_time_gtor().branch_time_variance(i);
  }

  report_time_instance(time_label,branch_time,mdl.tree(),os);
}


unsigned
subs_ml_model::
param_size(const PARAM_VIEW::index_t pv) const {

  return (get_time_gtor().param_size(pv)+
          get_rate_gtor().param_size(pv)+
          get_root_gtor().param_size(pv));
}


const char* const section_id = "subs_ml_model";

void
subs_ml_model::
store_state(std::ostream& os) const {

  os << section_id << " model_file_version " << MODEL_FILE_VERSION << "\n";
  os << section_id << " cmdline " << _cmdline << "\n";

  // report comex data first:
  os << section_id << " rate_gtor_model " << rate_gtor_model() << "\n";

  get_cat_manager().store_state(os);

  get_time_gtor().store_state(os);

  // report background values:
  const unsigned n_mapped_cats(get_cat_manager().get_assigned_cat_size()+1); // unassigned+assigned cats mapped from data
  const bool is_unassigned_cat(get_cat_manager().is_unassigned_cat());
  const unsigned n_states(state_size());
  const unsigned n_seqs(tree().n_leaves());
  const unsigned * const * const * obs(_obs.cat_seq_state_counts.val);

  os << section_id << " obs_state_mode 0\n";
  os << std::setprecision(SUBS_ML_PRINT_MODEL_PRECISION);
  for(unsigned k(0);k<n_mapped_cats;++k){
    if(k==UNASSIGNED_CAT_ID && ! is_unassigned_cat) continue;
    for(unsigned i(0);i<n_states;++i){
      os << section_id << " obs_";
      if(k==UNASSIGNED_CAT_ID) os << UNASSIGNED_CAT_LABEL;
      else                     os << "assigned_cat_" << k-1;
      os << "_seq_state_" << i;
      for(unsigned j(0);j<n_seqs;++j) os << " " << obs[k][j][i];
      os << "\n";
    }
  }

  get_rate_gtor().store_state(os);
  get_root_gtor().store_state(os);

  opt().store_state(os);
}




static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr)  NORETURN_TAG;

static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr){

  std::cerr << "ERROR:: -- Invalid model file format. --\n"
            << "ERROR:: expected_section: " << sid << "\n"
            << "ERROR:: section: " << tmpstr << "\n"
            << "\n";
  std::cerr << is.rdbuf();
  abort();
}




static
void
load_advance(std::istream& is,
             const char* const second = 0){

  std::string tmpstr;
  is >> tmpstr;
  if(tmpstr != section_id) load_die(is,section_id,tmpstr);
  is >> tmpstr;
  if(second && tmpstr != second) load_die(is,second,tmpstr);
}




void
subs_ml_model::
load_state(std::istream& is) {

  std::string dummy;

  {  // check file version
    unsigned tmpu;
    load_advance(is,"model_file_version");
    is >> tmpu;
    if(tmpu != MODEL_FILE_VERSION){
      std::cerr << "ERROR:: bad model file version number. expected: " << MODEL_FILE_VERSION << "\n";
      exit(EXIT_FAILURE);
    }
  }

  // throw away old cmd-line
  load_advance(is,"cmdline");
  clear_line(is);

  // pick up comex data first:
  //
  { // set polymorphic rate_gtor type:
    int tmpi;
    load_advance(is,"rate_gtor_model");
    is >> tmpi;
    _ragm=(static_cast<RATE_GTOR_MODEL::index_t>(tmpi));
    assert(_ragm != RATE_GTOR_MODEL::NONE);
  }
  _cm.reset(new cat_manager(is));

  _time_gtor.reset(new time_gtor(is));

  const unsigned n_mapped_cats(get_cat_manager().get_assigned_cat_size()+1); // unassigned+assigned cats mapped from data
  const bool is_unassigned_cat(get_cat_manager().is_unassigned_cat());
  const unsigned n_seqs(tree().n_leaves());
  const unsigned n_states(state_size());

  {
    int tmpi;
#ifdef NEXT_MODELFILE_VERSION
    load_advance(is,"obs_state_mode");
#else
    load_advance(is);
#endif
    is >> tmpi;
    assert(tmpi==0);
  }

  _obs.init(n_mapped_cats,n_seqs,n_states);
  unsigned*** obs(_obs.cat_seq_state_counts.val);
  for(unsigned k(0);k<n_mapped_cats;++k){
    if(k==UNASSIGNED_CAT_ID && ! is_unassigned_cat) {
      for(unsigned i(0);i<n_states;++i){
        for(unsigned j(0);j<n_seqs;++j){
          obs[k][j][i] = 0;
        }
      }
    } else {
      for(unsigned i(0);i<n_states;++i){
        load_advance(is);
        for(unsigned j(0);j<n_seqs;++j) is >> obs[k][j][i];
      }
    }
  }
  update_obs_distros();

  set_rate_gtor_model();
  get_rate_gtor().load_state(is);

  _root_gtor.reset(new root_gtor(is,root_gtor_sml_share(*this)));

  opt().load_state(is);

#ifdef DEBUG
  {
    const unsigned ps(param_size());
    smlfloat* pcopy(new smlfloat[ps]);
    param_state(pcopy);
    for(unsigned i(0);i<ps;++i){
      std::cerr << pcopy[i] << "\n";
      if( is_float_invalid(pcopy[i]) ){
        std::cerr << "ERROR:: subs_ml_model invalid input param: " << i << " " << pcopy[i] << "\n";
        abort();
      }
    }
    delete [] pcopy;
  }
#endif

}



void
subs_ml_model::
reset_param(const DENOVO_MODEL::index_t dm){
  get_time_gtor().reset_param(dm);
  get_rate_gtor().reset_param(dm);
  get_root_gtor().reset_param(dm);
}


void
subs_ml_model::
dump(std::ostream& os) const {
  get_time_gtor().dump(os);
  get_rate_gtor().dump("rate",os);
  //  pdistro_dump(get_root_gtor().distro(),get_root_gtor().state_size(),os,"p_root");
}




void
subs_ml_model::
report_time(std::ostream& os) const {

  report_subs_ml_model_time_instance(get_rate_gtor().rate_time_context_avg_bg_scale(),
                                     "context_bg_time",*this,os);

  report_subs_ml_model_time_instance(get_rate_gtor().rate_time_bg_scale(),
                                     "bg_time",*this,os);
}




void
subs_ml_model::
report(std::ostream& os) const {

  const unsigned ps(param_size(PARAM_VIEW::INDY_MIN_PARAM));
  os << "Free parameters: " << ps << "\n";
  os << "(subtract 1 time parameter for reversible models because a rooted tree is always used)\n";
  os << "\n";
  os << "root_distro: \n";
  {
    const unsigned ss(state_size());
    const prob_t* obsp(obs_state_distro());
    const root_gtor& rg(get_root_gtor());

    os << "State Size: " << ss << "\n";

    for(unsigned i(0);i<rg.root_cat_size();++i){
      const prob_t* root_pdf(rg.root_cat_distro(i));
      const smlfloat rdot(array_dot(root_pdf,obsp,ss)/
                          (array_norm2(root_pdf,ss)*array_norm2(obsp,ss)));
      os << "Category " << i << " dot to obs: " << rdot << "\n";
    }
    os << "\n";
  }
  /// \todo fix this hack:
  get_rate_gtor().report_root(get_root_gtor(),tree().root->label.c_str(),os);

  report_time(os);
  get_rate_gtor().report("ALL-BRANCH",os);
}



void
subs_ml_model::
set_obs_cat_seq_state_counts(unsigned const * const * const * c){
  const unsigned n_mapped_cats(get_cat_manager().get_assigned_cat_size()+1); // unassigned+assigned cats mapped from data
  const unsigned n_seqs(tree().n_leaves());
  const unsigned n_states(state_size());

  for(unsigned i(0);i<n_mapped_cats;++i){
    for(unsigned j(0);j<n_seqs;++j){
      for(unsigned k(0);k<n_states;++k){
        _obs.cat_seq_state_counts.val[i][j][k] = c[i][j][k];
      }
    }
  }

  update_obs_distros();
  get_rate_gtor().bg_pdistro_update();
}



void
subs_ml_model::
update_obs_distros(){
  const unsigned n_mapped_cats(get_cat_manager().get_assigned_cat_size()+1); // unassigned+assigned cats mapped from data
  const unsigned n_seqs(tree().n_leaves());
  const unsigned n_states(state_size());

  {
    prob_t* distro(_obs.state_distro.val);
    std::fill(distro,distro+n_states,0.);

    for(unsigned j(0);j<n_seqs;++j){
      prob_t* sdistro(_obs.seq_state_distro.val[j]);
      for(unsigned k(0);k<n_states;++k){
        sdistro[k]=0.;
        for(unsigned i(0);i<n_mapped_cats;++i){
          const prob_t val(static_cast<prob_t>(_obs.cat_seq_state_counts.val[i][j][k]));
          sdistro[k] += val;
          distro[k] += val;
        }
      }
      pdistro_norm_safe(sdistro,sdistro+n_states);
    }
    pdistro_norm_safe(distro,distro+n_states);
  }

  for(unsigned i(1);i<n_mapped_cats;++i){
    prob_t* cdistro(_obs.state_distro_assigned_cat.val[i-1]);
    std::fill(cdistro,cdistro+n_states,0.);
    for(unsigned j(0);j<n_seqs;++j){
      prob_t* scdistro(_obs.seq_state_distro_assigned_cat.val[i-1][j]);
      for(unsigned k(0);k<n_states;++k){
        const prob_t val(static_cast<prob_t>(_obs.cat_seq_state_counts.val[i][j][k]));
        scdistro[k] = val;
        cdistro[k] += val;
      }
      pdistro_norm_safe(scdistro,scdistro+n_states);
    }
    pdistro_norm_safe(cdistro,cdistro+n_states);
  }
}



void
subs_ml_model::
obs_mapped_cat_distro(prob_t* p) const {
  const unsigned n_mapped_cats(get_cat_manager().get_assigned_cat_size()+1); // unassigned+assigned cats mapped from data
  const unsigned n_seqs(tree().n_leaves());
  const unsigned n_states(state_size());

  std::fill(p,p+n_mapped_cats,0.);
  for(unsigned i(0);i<n_mapped_cats;++i){
    for(unsigned j(0);j<n_seqs;++j){
      for(unsigned k(0);k<n_states;++k){
        const unsigned val(_obs.cat_seq_state_counts.val[i][j][k]);
        p[i] += static_cast<prob_t>(val);
      }
    }
  }

  pdistro_norm_safe(p,p+n_mapped_cats);
}


