// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: subs_ml_model.h 802 2007-09-09 22:39:52Z ctsa $

/// \file

#ifndef __SUBS_ML_MODEL_H
#define __SUBS_ML_MODEL_H


#include "bi_tree_type.h"
#include "cat_manager.h"
#include "param_view.h"
#include "rate_gtor.h"
#include "rate_gtor_factory.h"
#include "root_gtor.h"
#include "simple_util.h"
#include "subs_ml_model_ci_info.h"
#include "subs_ml_model_comex.h"
#include "subs_ml_model_options.h"
#include "time_gtor.h"
#include "util/math/indy_random_var.h"

#include <cassert>
#include <cmath>

#include <iosfwd>
#include <string>
#include <vector>


/// \brief parameters and behavior of ml model
///
struct subs_ml_model {

  typedef subs_ml_model self_t;

  subs_ml_model(std::istream& is,
                const std::string& init_cmdline)
    : _cmdline(init_cmdline) {
    load_state(is);
  }

  subs_ml_model(const time_gtor_options& tgo,
                const RATE_GTOR_MODEL::index_t ragm,
                const rate_gtor_nuc_options& nopt,
                const rate_gtor_nscodon_options& copt,
                const C5_APPROX::index_t c5t,
                const ROOT_GTOR_MODEL::index_t rogm,
                const cat_manager_options& catman_opt,
                const std::string& init_cmdline,
                const DENOVO_MODEL::index_t dm)
    : _ragm(ragm),
      _cm(new cat_manager(catman_opt)),
      _time_gtor(new time_gtor(tgo)),
      _cmdline(init_cmdline) {

    assert(ragm != RATE_GTOR_MODEL::NONE);

    {
      const unsigned n_mapped_cats(get_cat_manager().get_assigned_cat_size()+1); // unassigned+assigned cats mapped from data
      const unsigned n_seqs(tree().n_leaves());
      const unsigned n_states(state_size());
      _obs.init(n_mapped_cats,n_seqs,n_states);

      std::fill(_obs.cat_seq_state_counts.val[0][0],_obs.cat_seq_state_counts.val[0][0]+n_mapped_cats*n_seqs*n_states,0);
      update_obs_distros();
    }

    set_rate_gtor_model(nopt,copt,c5t);
    _root_gtor.reset(new root_gtor(rogm,root_gtor_sml_share(*this)));

    reset_param(dm);
  }

  subs_ml_model(const self_t& s)
    :  _ragm(s._ragm),
       _cm(new cat_manager(*s._cm)), _time_gtor(new time_gtor(*s._time_gtor)),
       _rate_gtor(s._rate_gtor->clone(rate_gtor_sml_share(*this))),
       _root_gtor(new root_gtor(*s._root_gtor,root_gtor_sml_share(*this))),
       _opt(s._opt), _cmdline(s._cmdline), _obs(s._obs) {}

private:
  self_t& operator=(const self_t& rhs); // disable copy
public:

  RATE_GTOR_MODEL::index_t rate_gtor_model() const { return _ragm; }

  time_gtor& get_time_gtor(){ return *_time_gtor; }
  const time_gtor& get_time_gtor() const { return *_time_gtor; }

  rate_gtor_base& get_rate_gtor(){ return *_rate_gtor; }
  const rate_gtor_base& get_rate_gtor() const { return *_rate_gtor; }

  root_gtor& get_root_gtor(){ return *_root_gtor; }
  const root_gtor& get_root_gtor() const { return *_root_gtor; }

  subs_ml_model_options& opt() {return _opt;}
  const subs_ml_model_options& opt() const {return _opt;}

  const cat_manager& get_cat_manager() const { return *_cm; }

  const bi_tree_sml_type& tree() const { return get_time_gtor().tree(); }

  unsigned state_size() const { return RATE_GTOR_MODEL::state_size(rate_gtor_model()); }

  unsigned param_size(const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL) const;

  unsigned submodel_size() const { return get_rate_gtor().submodel_size(); }

  unsigned submodel_state_size(unsigned s) const { return get_rate_gtor().submodel_state_size(s); }

  template <typename OutputIterator>
  OutputIterator
  param_state(OutputIterator x,
              const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL) const;

  template <typename InputIterator>
  InputIterator
  set_param_state(InputIterator x,
                  const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL,
                  const bool is_secondary_update = true);

  template <typename OutputIterator>
  OutputIterator
  is_train_param(OutputIterator i) const;

  template <typename InputIterator>
  InputIterator
  set_is_train_param(InputIterator i);

  void store_state(std::ostream& os) const;

  smlfloat param_penalty() const {
    return get_rate_gtor().param_penalty()+get_time_gtor().param_penalty();
  }

  /// \brief quick debugging output of model parameters
  void dump(std::ostream& os) const;

  /// \brief report information processed from model parameters
  void report(std::ostream& os) const;

  /// \brief time subsection of report
  void report_time(std::ostream& os) const;

  const std::string& cmdline() const { return _cmdline; }

#if 0
  unsigned cat_size() const { return get_cat_manager().cat_size(); }

  unsigned site_cat_size() const { return get_cat_manager().site_cat_size(); }

  unsigned group_cat_size() const { return get_cat_manager().group_cat_size(); }
#endif

  template <typename InputIterator>
  InputIterator
  attach_ci(InputIterator ci){
    ci = get_time_gtor().attach_ci(ci);
    ci = get_rate_gtor().attach_ci(ci);
    ci = get_root_gtor().attach_ci(ci);
    return ci;
  }

  void set_obs_cat_seq_state_counts(const unsigned * const * const * c);

  void obs_mapped_cat_distro(prob_t* p) const;

  const prob_t * const * obs_seq_state_distro() const { return _obs.seq_state_distro.val; }

  const prob_t * obs_state_distro() const { return _obs.state_distro.val; }

  const prob_t * const * obs_seq_state_distro_assigned_cat(const unsigned c) const {
    assert(c<get_cat_manager().get_assigned_cat_size());
    return _obs.seq_state_distro_assigned_cat.val[c];
  }

  const prob_t * obs_state_distro_assigned_cat(const unsigned c) const {
    assert(c<get_cat_manager().get_assigned_cat_size());
    return _obs.state_distro_assigned_cat.val[c];
  }

private:
  void load_state(std::istream& is);

  void set_rate_gtor_model(const rate_gtor_nuc_options& nopt = rate_gtor_nuc_options(),
                           const rate_gtor_nscodon_options& copt = rate_gtor_nscodon_options(),
                           const C5_APPROX::index_t c5t = C5_APPROX::NONE){
    _rate_gtor.reset(rate_gtor_factory(rate_gtor_sml_share(*this),nopt,copt,c5t));
  }

  void reset_param(const DENOVO_MODEL::index_t dm);

  void check_param() const;

  void update_obs_distros();

  //////
  RATE_GTOR_MODEL::index_t _ragm;
  std::auto_ptr<cat_manager> _cm;

  std::auto_ptr<time_gtor> _time_gtor;
  std::auto_ptr<rate_gtor_base> _rate_gtor;
  std::auto_ptr<root_gtor> _root_gtor;

  subs_ml_model_options _opt;

  std::string _cmdline;

  struct obs_info {
    void init(const unsigned n_mapped_cats,
              const unsigned n_seqs,
              const unsigned n_states){
      cat_seq_state_counts.init(n_mapped_cats,n_seqs,n_states);
      seq_state_distro.init(n_seqs,n_states);
      state_distro.init(n_states);
      seq_state_distro_assigned_cat.init(n_mapped_cats-1,n_seqs,n_states);
      state_distro_assigned_cat.init(n_mapped_cats-1,n_states);
    }

    simple_init_matrix3d<unsigned> cat_seq_state_counts;
    simple_init_matrix<prob_t> seq_state_distro;
    simple_init_array<prob_t> state_distro;
    simple_init_matrix3d<prob_t> seq_state_distro_assigned_cat;
    simple_init_matrix<prob_t> state_distro_assigned_cat;
  };

  obs_info _obs;
};


#include "subs_ml_model.hh"

#endif
