// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: subs_ml_model_util.cc 817 2007-09-18 01:06:28Z ctsa $

/// \file

#include "cat_info.h"
#include "catmap_util.h"
#include "rate_gtor_nuc_base.h"
#include "rate_gtor_nscodon_base.h"
#include "rate_gtor_util.h"
#include "simple_util.h"
#include "site_data_stats.h"
#include "subs_ml_model_util.h"



void
init_subs_ml_model_trained_data_dependencies(subs_ml_model& mdl,
                                             const site_data& sd,
                                             const std::vector<bool>& is_assigned_cats){

  const unsigned n_states(mdl.state_size());
  const unsigned n_model_assigned_cats(mdl.get_cat_manager().get_assigned_cat_size());

  if(n_model_assigned_cats==0){
    simple_array<prob_t> distro(n_states);
    site_data_2_state_pdf(sd,n_states,distro.val);

    mdl.get_root_gtor().set_distro(distro.val);

  } else {
    const unsigned n_data_cats(sd.max_cat()+1);

    // catlup is mapping from data_cats to model_cats:
    std::vector<unsigned> catlup;
    catmap_to_catlup(sd,n_model_assigned_cats,mdl.get_cat_manager().data_catmap(),catlup);

    simple_array<prob_t> model_assigned_cat_prior(n_model_assigned_cats);

    { // get data cat counts, map this to model cat counts, convert counts to distro

      std::vector<unsigned> data_cat_count(n_data_cats,0);
      site_data_2_cat_count(sd,n_data_cats,data_cat_count);

      array_zero(model_assigned_cat_prior.val,n_model_assigned_cats);
      for(unsigned i(0);i<n_data_cats;++i){
        if(catlup[i]==UNASSIGNED_CAT_ID) continue;
        if(catlup[i]>n_model_assigned_cats) continue;
        model_assigned_cat_prior.val[catlup[i]-1] += data_cat_count[i];
      }

      pdistro_norm(model_assigned_cat_prior.val,model_assigned_cat_prior.val+n_model_assigned_cats);
    }


    if(is_assigned_cats[SITE_CAT_TYPE::MUT_RATE] ||
       is_assigned_cats[SITE_CAT_TYPE::MUT_MODEL]){
      rate_gtor_nuc_base* rg_ptr(dynamic_cast<rate_gtor_nuc_base*>(&mdl.get_rate_gtor()));
      if(! rg_ptr) die("site model incompatible with mutation categories");
      if(is_assigned_cats[SITE_CAT_TYPE::MUT_RATE]){
        rg_ptr->set_site_rate_cats_prob(model_assigned_cat_prior.val);
      } else {
        rg_ptr->set_site_mutation_model_cats_prob(model_assigned_cat_prior.val);
      }

    } else if(is_assigned_cats[SITE_CAT_TYPE::SEL_STRENGTH] ||
              is_assigned_cats[SITE_CAT_TYPE::SEL_MATRIX]){
      rate_gtor_nscodon_base* rg_ptr(dynamic_cast<rate_gtor_nscodon_base*>(&mdl.get_rate_gtor()));
      if(! rg_ptr) die("site_model incompatible with selection categories");

      if(is_assigned_cats[SITE_CAT_TYPE::SEL_STRENGTH]){
        rg_ptr->set_site_select_cats_prob(model_assigned_cat_prior.val);
      } else {
        rg_ptr->set_site_select_matrix_cats_prob(model_assigned_cat_prior.val);
      }
    }

    {
      std::vector<unsigned> data_cat_state_count;
      site_data_2_cat_state_count(sd,n_data_cats,n_states,data_cat_state_count);

      const unsigned mcsd_size(n_model_assigned_cats*n_states);
      std::vector<prob_t> model_cat_state_distro(mcsd_size,0.);

      for(unsigned i(0);i<n_data_cats;++i){
        if(catlup[i]==UNASSIGNED_CAT_ID) continue;
        if(catlup[i]>n_model_assigned_cats) continue;
        const unsigned model_state(catlup[i]-1);
        for(unsigned j(0);j<n_states;++j){
          model_cat_state_distro[model_state*n_states+j] += data_cat_state_count[i*n_states+j];
        }
      }

      for(unsigned i(0);i<n_model_assigned_cats;++i){
        pdistro_norm(model_cat_state_distro.begin()+i*n_states,model_cat_state_distro.begin()+(i+1)*n_states);
      }

      mdl.get_root_gtor().set_all_root_cat_distro(model_cat_state_distro.begin());
    }
  }
}



void
init_subs_ml_model_untrained_data_dependencies(subs_ml_model& mdl,
                                               const site_data& sd){

  const unsigned n_states(mdl.state_size());
  const unsigned n_data_cats(sd.max_cat()+1);
  std::vector<unsigned> data_cat_state_count;
  site_data_2_cat_state_count(sd,n_data_cats,n_states,data_cat_state_count);

  const unsigned n_model_assigned_cats(mdl.get_cat_manager().get_assigned_cat_size());

  // catlup is mapping from data_cats to model_cats:
  std::vector<unsigned> catlup;
  catmap_to_catlup(sd,n_model_assigned_cats,mdl.get_cat_manager().data_catmap(),catlup);

  const unsigned n_seqs(sd.n_taxa());
  const unsigned n_model_mapped_cats(n_model_assigned_cats+1); // unassigned+assigned mappings from data_cats
  simple_matrix3d<unsigned> sd_cat_seq_state_counts(n_model_mapped_cats,n_seqs,n_states);

  std::fill(sd_cat_seq_state_counts.val[0][0],
            sd_cat_seq_state_counts.val[0][0]+n_model_mapped_cats*n_seqs*n_states,
            0);

  for(unsigned i(0);i<n_data_cats;++i){
    if(catlup[i]>n_model_assigned_cats) continue;
    site_data_2_taxa_state_count(sd,n_states,sd_cat_seq_state_counts.val[catlup[i]][0],true,true,i);
  }

  // map from sd seqs to tree leaves:
  std::vector<unsigned> sd_org_tree_order(n_seqs);
  for(unsigned j(0);j<n_seqs;++j){
    sd_org_tree_order[j] = sd.taxid.getid(mdl.tree().leaf[j]->label);
  }

  // remap sequence order
  simple_matrix<unsigned> ptmp(n_seqs,n_states);
  for(unsigned i(0);i<n_model_mapped_cats;++i){
    unsigned** si(sd_cat_seq_state_counts.val[i]);
    for(unsigned j(0);j<n_seqs;++j){
      for(unsigned k(0);k<n_states;++k){
        ptmp.val[j][k] = si[j][k];
      }
    }
    for(unsigned j(0);j<n_seqs;++j){
      for(unsigned k(0);k<n_states;++k){
        si[j][k] = ptmp.val[sd_org_tree_order[j]][k];
      }
    }
  }

  mdl.set_obs_cat_seq_state_counts(sd_cat_seq_state_counts.val);
}




void
norm_model(subs_ml_model& mdl){

  const unsigned n_states(mdl.state_size());
  const unsigned n_assigned_cats(mdl.get_cat_manager().get_assigned_cat_size());
  const bool is_unassigned_cat(mdl.get_cat_manager().is_unassigned_cat());
  //  const unsigned n_root_cats();

  if(n_assigned_cats != 0 || ! is_unassigned_cat){
    /// \todo fix norm_model to work with assigned cats
    die("norm_model has not yet been updated for assigned categories");
  }

#if 0
  simple_matrix<smlfloat> assigned_cat_distro(n_assigned_cats,n_states);
  for(unsigned i(0);i<n_assigned_cats;++i){
    array_zero(assigned_cat_distro.val[i],n_states);
  }
#endif

  std::vector<smlfloat> unassigned_distro(n_states,0.);

  {
    simple_array<smlfloat> site_cat_distro(n_states);

    const unsigned n_group_cats(mdl.get_rate_gtor().category_size_group());

    std::vector<prob_t> group_cat_prior_pdistro(n_group_cats);
    mdl.get_rate_gtor().get_group_category_pdistro(group_cat_prior_pdistro.begin());

    const unsigned n_site_cats(mdl.get_rate_gtor().category_size_site());

    std::vector<prob_t> site_cat_prior_pdistro(n_site_cats);
    mdl.get_rate_gtor().get_site_category_pdistro(site_cat_prior_pdistro.begin());

    for(unsigned gc(0);gc<n_group_cats;++gc){
      for(unsigned sc(0);sc<n_site_cats;++sc){
        prob_t cat_prob(site_cat_prior_pdistro[sc]*group_cat_prior_pdistro[gc]);
        get_stationary_pdistro(site_cat_distro.val,mdl.get_rate_gtor(),rates_func_options(sc,gc));
        pdistro_check(site_cat_distro.val,n_states,SUBS_ML_PTOL);

        //        const unsigned root_cat(mdl.get_cat_manager().get_typed_cat_no(SITE_CAT_TYPE::ROOT,sc));
        for(unsigned j(0);j<n_states;++j){
          const prob_t val(site_cat_distro.val[j]*cat_prob);
          unassigned_distro[j] += val;
#if 0
          root_distro.val[root_cat][j] += val;
#endif
        }
      }
    }
  }

  pdistro_norm(unassigned_distro.begin(),unassigned_distro.end());

  {
    /// \todo find a better (exact) way to handle this
    static const unsigned fake_counts(std::numeric_limits<unsigned>::max()-(1000));

    const unsigned n_mapped_cats(mdl.get_cat_manager().get_assigned_cat_size()+1); // unassigned+assigned cats mapped from data
    const unsigned n_seqs(mdl.tree().n_leaves());

    simple_matrix3d<unsigned> unassigned_counts(n_mapped_cats,n_seqs,n_states);

    for(unsigned k(0);k<n_states;++k){
      const unsigned val(static_cast<unsigned>(unassigned_distro[k]*fake_counts));
      for(unsigned i(0);i<n_mapped_cats;++i){
        for(unsigned j(0);j<n_seqs;++j){
          unassigned_counts.val[i][j][k] = val;
        }
      }
    }

    mdl.set_obs_cat_seq_state_counts(unassigned_counts.val);
    mdl.get_root_gtor().set_all_root_cat_distro(unassigned_distro.begin());
  }

#if 0
  for(unsigned i(0);i<n_root_cats;++i){
    pdistro_norm(root_distro.val[i],root_distro.val[i]+n_states);
  }
  mdl.get_root_gtor().set_all_root_cat_distro(root_distro.val[0]);
#endif
}
