// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: time_gtor.cc 802 2007-09-09 22:39:52Z ctsa $

/// \file

#include "subs_ml_print_util.h"
#include "substk_exception.h"
#include "time_gtor.h"
#include "util/general/metatags.h"
#include "util/math/random_util.h"
#include "util/math/test_float.h"

#include <iomanip>
#include <iostream>
#include <sstream>


void
time_gtor::
adjust_time(smlfloat& time) {
  static const smlfloat MAX_BRANCH_TIME(1.5);
  if(time > MAX_BRANCH_TIME) {
    _time_penalty += (time - MAX_BRANCH_TIME);
    std::cerr  << "time penalty triggered, converted from: " << time << " to " << MAX_BRANCH_TIME << "\n";
    time = MAX_BRANCH_TIME;
  } else if(time<=ZERO_BRANCH_TIME_THRESHOLD) {
    time = 0.;
  }
}


void
time_gtor::
check_pseudoparam_num(unsigned param_no) const {
  if(param_no>=pseudo_param_size()){
    die("Bad pseudoparam number");
  }
}


const char* const section_id = "time_gtor";

void
time_gtor::
store_state(std::ostream& os) const {

  os << section_id << " tree " << tree() << "\n";

  const unsigned ps(param_size());

  os << std::setprecision(SUBS_ML_PRINT_MODEL_PRECISION);
  for(unsigned i(0);i<ps;++i){
    os << section_id << " time " << _tree.branch_label(i) << " " <<
      _is_train_param[i] << " " <<
      branch_time(i) << "\n";
  }
}




static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr)  NORETURN_TAG;

static
void
load_die(std::istream& is,
         const char* const sid,
         const std::string& tmpstr){

  std::cerr << "ERROR:: -- Invalid model file format. --\n"
            << "ERROR:: expected_section: " << sid << "\n"
            << "ERROR:: section: " << tmpstr << "\n"
            << "\n";
  std::cerr << is.rdbuf();
  abort();
}



static
void
load_advance(std::istream& is,
             const char* const second = 0){

  std::string tmpstr;
  is >> tmpstr;
  if(tmpstr != section_id) load_die(is,section_id,tmpstr);
  is >> tmpstr;
  if(second && tmpstr != second) load_die(is,second,tmpstr);
}



void
time_gtor::
load_state(std::istream& is) {

  std::string dummy;

  load_advance(is,"tree");
  is >> dummy;
  _tree.init(dummy.c_str());
  _tree.clear_lengths();

  const unsigned ps(param_size());
  _param.resize(ps);
  _is_train_param.resize(ps);

  for(unsigned i(0);i<ps;++i) {
    load_advance(is,"time");
    is >> dummy;
    const unsigned bi(tree().get_branch_index(dummy));
    bool is_t;
    is >> is_t >> _param[bi];
    _is_train_param[bi] = is_t;
  }

  check_param();
}




void
time_gtor::
reset_param(const DENOVO_MODEL::index_t dm){

  if(dm == DENOVO_MODEL::RANDOM){
    static const smlfloat LN_TEN = std::log(smlfloat(10.));
    static const smlfloat log10_time_min = -3.;
    static const smlfloat log10_time_max = -0.3;

    static const smlfloat timerange = log10_time_max-log10_time_min;

    for(unsigned i(0);i<branch_size();++i){
      if(_is_train_param[i]){
        const smlfloat rand_log10 = random_uniform()*timerange+log10_time_min;
        _param[i] = std::exp(rand_log10*LN_TEN);
      }
    }

    _tree.clear_lengths();
  } else if(dm == DENOVO_MODEL::START){
    static const smlfloat start_time(0.02);
    for(unsigned i(0);i<branch_size();++i){
      if(_is_train_param[i]){
        _param[i] = start_time;
      }
    }

    // pull init times from tree, if available:
    for(unsigned i(0);i<branch_size();++i){
      if(tree().is_branch_length_set(i)){
        _param[i] = tree().branch_length(i);
      }
    }

    _tree.clear_lengths(false);

  } else { die("unknown denovo_model"); }
}



void
time_gtor::
check_param() const {
  for(unsigned i(0);i<param_size();++i) {
    if(is_float_invalid(_param[i])){
      std::ostringstream oss;
      oss << "Invalid model parameter: " << i << " " << _param[i] << "\n";
      std::cerr << oss.str();
      throw substk_exception(oss.str().c_str());
    }
  }

  for(unsigned i(0);i<branch_size();++i){
    if(branch_time(i)<0.){
      std::ostringstream oss;
      oss << "Negative branch time parameter: " << i << " " << branch_time(i) << "\n";
      std::cerr << oss.str();
      throw substk_exception(oss.str().c_str());
    }
  }
}



void
time_gtor::
dump(std::ostream& os) const {
  for(unsigned i(0);i<branch_size();++i){
    os << branch_label(i) << "\t" << branch_time(i) << "\n";
  }
}



void 
time_gtor::
set_reversible_time_param(){

  const bi_tree_node* c1(tree().root->child1);
  const bi_tree_node* c2(tree().root->child2);
  if(c1 == 0 || c2 == 0) return;

  // check if either branch is already locked:
  const unsigned bi1(tree().get_branch_index(c1));

  const unsigned bi2(tree().get_branch_index(c2));
  if( _is_train_param[bi1]==false || _is_train_param[bi2]==false) return;

  // lock child1, unless it's a leaf node:
  unsigned bix(bi1);
  if(c1->child1 == 0) bix=bi2;

  _param[bix]=0.;
  _is_train_param[bix]=false;
}
