// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: time_gtor.h 802 2007-09-09 22:39:52Z ctsa $

/// \file

#ifndef __TIME_GTOR_H
#define __TIME_GTOR_H


#include "bi_tree_type.h"
#include "denovo_types.h"
#include "param_view.h"
#include "subs_ml_types.h"
#include "util/math/indy_random_var.h"
#include "util/general/die.h"

#include <string>
#include <vector>


struct time_gtor_options {

  time_gtor_options() :
    tree_buffer(), is_reversible(false) {}

  std::string tree_buffer;
  bool is_reversible;
};



struct time_gtor {

  typedef time_gtor self_t;

  explicit
  time_gtor(std::istream& is)
    : _time_penalty(0), _tree(), _param(), _is_train_param(), _param_variance(){
    load_state(is);
  }

  explicit
  time_gtor(const time_gtor_options& tgo)
    : _time_penalty(0.), _tree(tgo.tree_buffer.c_str()), _param(_tree.n_branches(),0.),
      _is_train_param(_tree.n_branches(),true), _param_variance() {
    if(tgo.is_reversible) set_reversible_time_param();
  }

private:
  self_t& operator=(const self_t& rhs);
public:

  unsigned param_size(const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL) const {
    switch(pv){
    case PARAM_VIEW::ALL :
      return tree().n_branches();
    case PARAM_VIEW::MIN_PARAM :
      return  param_size_trainable() + pseudo_param_size();
    case PARAM_VIEW::TRAINABLE:
    case PARAM_VIEW::INDY_MIN_PARAM :
      return param_size_trainable();
    default: die("unknown param view");
    }
  }

  unsigned branch_size() const { return param_size(); }

  const smlfloat& branch_time(int i) const {
    assert(i<static_cast<int>(branch_size())); return _param[i];
  }

  std::vector<smlfloat>::const_iterator branch_times() const {
    return _param.begin();
  }

  const irv_t<smlfloat> branch_time_variance(int i) const {
    assert(i<static_cast<int>(branch_size()));
    return irv_t<smlfloat>(_param[i],param_variance(i));
  }

  const std::string& branch_label(int i) const {
    assert(i<static_cast<int>(branch_size())); return tree().branch_label(i);
  }

  template <typename OutputIterator>
  OutputIterator
  param_state(OutputIterator x,
              const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL) const;

  template <typename InputIterator>
  InputIterator
  set_param_state(InputIterator x,
                  const PARAM_VIEW::index_t pv = PARAM_VIEW::ALL);

  template <typename FloatType>
  void set_pseudo_param(const unsigned param_no, const FloatType val);

  template <typename OutputIterator>
  OutputIterator
  is_train_param(OutputIterator i) const;

  template <typename InputIterator>
  InputIterator
  set_is_train_param(InputIterator i);

  void store_state(std::ostream& os) const;

  void load_state(std::istream& is);

  /// \brief quick debugging output of model parameters
  void dump(std::ostream& os) const;

#if 0
  /// \brief report information processed from model parameters
  void report(std::ostream& os) const;

  /// \brief time subsection of report
  void report_time(std::ostream& os) const;
#endif
  template <typename InputIterator>
  InputIterator
  attach_ci(InputIterator ci){
    const unsigned ps(param_size());
    _param_variance.clear();
    _param_variance.resize(ps,0.);
    for(unsigned i(0);i<ps;++i){
      if(_is_train_param[i]){
        _param_variance[i] = ci->variance;
        ci++;
      }
    }

    return ci;
  }

  const smlfloat& param_penalty() const { return _time_penalty; }

  void reset_param(const DENOVO_MODEL::index_t dm);

  const bi_tree_sml_type& tree() const { return _tree; }

  // client-code can decide what to do about this, it's
  // a perfectly valid state for the tree to be in:
  //
  bool is_zero_time() const {
    smlfloat branch_sum(0.);
    for(unsigned i(0);i<branch_size();++i){
      branch_sum += branch_time(i);
    }
    return (branch_sum <= 0.);
  }

private:

  /// adjusts time parameters for a reversible model
  ///
  void set_reversible_time_param();

  unsigned param_size_trainable() const {
    const unsigned ps(param_size());
    unsigned bst(0);
    for(unsigned i(0);i<ps;++i){ if(_is_train_param[i]) bst++; }
    return bst;
  }

  unsigned pseudo_param_size() const { return 1; } // pseudoparam to scale all times together

  void check_param() const;

  smlfloat
  param_variance(const unsigned i) const {
    if(_param_variance.size()>i){
      return _param_variance[i];
    } else {
      return 0.;
    }
  }

  void adjust_time(smlfloat& time);

  void check_pseudoparam_num(unsigned param_no) const;

  //////
  smlfloat _time_penalty;

  bi_tree_sml_type _tree;

  std::vector<smlfloat> _param;
  std::vector<bool> _is_train_param;
  std::vector<smlfloat> _param_variance;
};

#include "time_gtor.hh"

#endif
