// -*- mode: c++; indent-tabs-mode: nil; -*-
//
//
// CodeAxe : phylogenetic analysis and simulation tools
//
//   http://www.phrap.org
//
//
// Copyright 2007 Christopher T Saunders (ctsa@u.washington.edu)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of the
// License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
// 02110-1301, USA.
//
//

// $Id: time_gtor.hh 741 2007-08-14 02:39:19Z ctsa $

/// \file

#include "subs_prob_util.h"

#include <cmath>

template <typename OutputIterator>
OutputIterator
time_gtor::
is_train_param(OutputIterator x) const {
  const unsigned ps(param_size());
  for(unsigned i(0);i<ps;++i) {
    *x = _is_train_param[i];
    ++x;
  }
  return x;
}




template <typename InputIterator>
InputIterator
time_gtor::
set_is_train_param(InputIterator x) {
  const unsigned ps(param_size());
  for(unsigned i(0);i<ps;++i) {
    _is_train_param[i] = *x;
    ++x;
  }

  return x;
}




template <typename OutputIterator>
OutputIterator
time_gtor::
param_state(OutputIterator x,
            const PARAM_VIEW::index_t pv) const {

  const bool is_min_transform(pv == PARAM_VIEW::MIN_PARAM ||
                              pv == PARAM_VIEW::INDY_MIN_PARAM);

  const bool is_trainable(is_min_transform ||
                          pv == PARAM_VIEW::TRAINABLE);

  if(pv == PARAM_VIEW::MIN_PARAM){
    const unsigned pps(pseudo_param_size());
    for(unsigned i(0);i<pps;++i) { *x = 1.; ++x; }
  }

  const unsigned ps(param_size());
  for(unsigned i(0);i<ps;++i) {
    if(is_trainable){
      if(_is_train_param[i]){
        smlfloat time_tmp = _param[i];
        if(is_min_transform){
          // time is presented to the minimizer in a more robust format,
          // meant to prevent "popups" to very high time values

          // adjust down from zero thresh slightly so that we can be
          // sure that x = exp(log(x)) is still <= zero thresh
          static const smlfloat fudge_down(0.99);
          time_tmp = std::max(time_tmp,ZERO_BRANCH_TIME_THRESHOLD*fudge_down);
          if(time_tmp>1.) time_tmp *= time_tmp;
          time_tmp = std::log(time_tmp);
        }
        *x = time_tmp; ++x;
      }
    } else {
      *x = _param[i];
      ++x;
    }
  }

  return x;
}




template <typename InputIterator>
InputIterator
time_gtor::
set_param_state(InputIterator x,
                const PARAM_VIEW::index_t pv) {

  typedef typename std::iterator_traits<InputIterator>::value_type InputT;

  const bool is_min_transform(pv == PARAM_VIEW::MIN_PARAM ||
                              pv == PARAM_VIEW::INDY_MIN_PARAM);

  const bool is_trainable(is_min_transform ||
                          pv == PARAM_VIEW::TRAINABLE);

  // 1) store pseudoparams
  const unsigned pps(pseudo_param_size());
  std::vector<InputT> p;
  if(pv == PARAM_VIEW::MIN_PARAM){
    p.resize(pps);
    for(unsigned i(0);i<pps;++i){ p[i] = *x; ++x; }
  }

  _time_penalty=0.;

  const unsigned ps(param_size());
  for(unsigned i(0);i<param_size();++i){
    if(is_trainable){
      if(_is_train_param[i]){
        smlfloat time_tmp = *x; ++x;
        if(is_min_transform){
          time_tmp = std::exp(time_tmp);
          if(time_tmp>1.) time_tmp = std::sqrt(time_tmp);
        }
        _param[i] = time_tmp;
      }
    } else {
      _param[i] = *x; ++x;
    }
  }

  // go back and handle pseudoparams
  if(pv == PARAM_VIEW::MIN_PARAM){
    for(unsigned i(0);i<pps;++i) set_pseudo_param(i,p[i]);
  }

  // put in max and zero cutoffs after the pseudoparam has a chance to scale times
  for(unsigned i(0);i<ps;++i){
    if(is_min_transform){
      if(_is_train_param[i]){
        adjust_time(_param[i]);
      }
    }
  }

  check_param();

  return x;
}




template <typename FloatType>
void
time_gtor::
set_pseudo_param(const unsigned param_no,
                 const FloatType val) {

  check_pseudoparam_num(param_no);

  // pseudoparam 1: scale all trainable times as a group:
  const FloatType safe_val(std::fabs(val));
  const unsigned ps(param_size());
  for(unsigned i(0);i<ps;++i) {
    if(_is_train_param[i]) {
      _param[i] *= safe_val;
    }
  }
}
